# QuantC

QuantC is a dialect of the C language, with extensions to allow for writing
programs for quantum algorithms emphasizing separation of concern. The
additional syntax specific to writing quantum algorithms are limited to
functions separate from regular C code, similar to how code is written for GPUs
when using CUDA or OpenGL. The intent to make it easier to generate working
software that requires interoperation between a classical and quantum computer.

To create a section of code intended to be ran on a quantum computer, you can
add the `quantum` storage specifier to a function.

```c
quantum int myFunction() {
    ...
}
```

Inside these quantum functions, you can use the `quantum` storage specifier on
variables to access qubits. These quantum variables can be operated on with
newly added quantum operations.

```c
quantum int myQubits = 0;

// Hadamard
%H myQubits

// Measure
%M myQubits
```

The following are the standard quantum operations you can perform on quantum
variables.

```c
%I myQubits;             // Identity (No change)
%X myQubits;             // X-Gate (Bitwise NOT extended to qubits)
%Y myQubits;             // Y-gate
%Z myQubits;             // Z-Gate
%H myQubits;             // Hadarmard
%PHASE(angle) myQubits;  // Phase
%S myQubits;             // S-Gate
%T myQubits;             // T-Gate
%RX(angle) myQubits;     // Rotate-X Gate
%RY(angle) myQubits;     // Rotate-Y Gate
%RZ(angle) myQubits;     // Rotate-Z Gate
%M myQubits;             // Measurement

myQubits ^ myOtherQubits          // CNOT (Bitwise XOR extended to qubits)
myQubits <P>(angle) myOtherQubits // Parametric Swap
myQubits <> myOtherQubits         // Classical Swap
myQubits <I> myOtherQubits        // Imaginary Swap
```

These quantum operations (excluding measurement) can have modifiers applied to
them to change there effect. The following are the currently available
modifiers.

```c
%H myQubits;                    // Regular Hadamard gate
~>%H myQubits;                  // Inverse of the Hadamard gate
%H+C(myOtherQubits) myQubits;   // Hadamard gate controlled by other qubits
```

This project is an compiler for the QuantC language. The goal is to have the
compiler generate object files and/or executables like regular C code in
addition to assembly files for quantum computers/simulators, such as Quil.
Calls to code written in the quantum assembly files are converted in the
executable/object files to instructions to transmit sections of said files to a
target quantum computer/simulator.

# Dependecies
* gcc (9.4.0+)
* CMake (3.23.2+)
* LLVM (10.0.0+)
* libcurl (7.7.1+)
* libjsoncpp-dev (1.7.4+)
* ForestSDK (2.23.0+)

# Components of the Project
This projects consists of the following components.

* Scanner (Lexer)
* Parser
* Abstract Syntax Tree (AST), with methods for
    * Semantics Checking
    * IR and Quantum Assembly Generation (LLVM IR & Quil)
    * Optimizing and Code Generation (LLVM)

# Project Structure
```
tests/
    full/
    base_c/
        tags/
        src/       # Numerous small C programs
        expected/  # Expected otuputs of programs in src/ if any
        LICENSE
code/
    include/
    quil-snippets/
    src/
        util/
        token/
        symbol_table/
        ast/
    test_runner/
CMakeLists-QPU.txt
CMakeLists.txt
main.cpp
qpu_driver.cpp
qpu_driver.h
scanner_shim.h
quant_c.lex
quant_c.y
```

The Flex scanner and Bison parser specifications are included in the
`quant_c.lex` and `quant_c.y` files, respectively. A simple shim
for the scanner generated by Flex that is needed to operate with the Bison
parser is defined in `scanner_shim.h`. Code to test different phases of the compiler is in `test_runner`. Sample
programs to test are included in the `tests/` directory.
Code used by the parser and scanner to build tokens and AST nodes is included
in the `src` directory.  

# How to Build
This project uses CMake to generate build files and build the compiler.

To generate the build files, run the following command inside the `code/` directory.

```cmake -B build/ -S . -DCMAKE_EXPORT_COMPILE_COMMANDS=1```

After generating the build files, run the following command to generate all of
the executables and binaries for this project.

```cmake --build build/```

You can also specify specific targets to build by running the previous command
with the `--target` flag.

```cmake --build build/ --target <target>```

Current relevant targets include

* `all` (the default when target not specified)
* `clean`
* `test_lexer`
* `test_parser`
* `test_codegen`
* `QuantC-compile` (The main program for the full compiler)
* `qpu_driver`     (The quantum communications library)

`clean` removes all the binary and other generated files created when building
with CMake. `test_lexer` builds an executable to test the scanner built using
Flex. Likewise, `test_parser` builds an executable to test the parser built
using Bison.

# How to Run

The executables currently generated by the build scripts are to test certain
components of the compiler, specifically the scanner and parser. Some test
files written in both pure C and the extended QuantC language are included in
the `tests/` directory.

To run one of the target programs, run the following command.

```
bin/<target_name> path/to/program
```

# Notes About Testing

The code currently assumes a specific test environment for the quantum server. The ForestSDK provides a quantum virtual machine (QVM). After installing, run the following command to setup the server on a local port.

```qvm -S > /dev/null 2>&1 &```

If desired, you can change where `stdout` directs to see the any output or error messages from the QVM.
