#include <iostream>
#include <fstream>
#include <filesystem>
#include <cstdlib>
#include "token/token.hpp"
#include "ast/ast.hpp"
#include "scanner_shim.hpp"
#include "quant_c_parse.hpp"

#define MIN_ARGC 2
#define EOF_TOKEN 0

using namespace quant_c_token;
using namespace quant_c_compiler;
using namespace quant_c_ast;

int main(int argc, char *argv[]) {

    if (argc < MIN_ARGC) {
        std::cerr << "Expected one argument; "
                  << "the name of a file to generate code for"
                  << std::endl;
        return -1;
    }

    std::filesystem::path tmpPath =
                            std::filesystem::temp_directory_path() / "quantc";
    if (!std::filesystem::exists(tmpPath)) {
        try {
            bool created = std::filesystem::create_directory(tmpPath);
            if (!created) {
                std::cerr << "Failed to create temporary directory" << std::endl;
                return -1;
            }
        }
        catch(...) {
            std::cerr << "Failed to create temporary directory" << std::endl;
            return -1;
        }
    }
    std::string tmp = tmpPath.string();

    // Preprocess the program before parsing
    std::string cd = std::filesystem::current_path();
    std::filesystem::path original_path(argv[1]);
    std::string file_name = std::string(original_path.stem());
    std::string preprocess_command = "gcc -E -P ";
    std::string preprocessed_file = tmp + "/" + file_name + ".i";

    preprocess_command += std::string(argv[1]) + " -o " + preprocessed_file;

    if (system(preprocess_command.c_str()) != EXIT_SUCCESS) {
        std::cout << "Error preprocessing file";
        return -1;
    }

    // Open the file to scan
    std::cout << "Reading file\n";
    std::ifstream quant_c_file;
    try {
        quant_c_file.open(preprocessed_file);
    } catch (...) {
        std::cerr << "Unable to open file \"" << preprocessed_file << "\"\n";
        return -1;
    }

    // Create the parser and parse the file
    std::cout << "Creating scanner\n";
    Scanner test_scanner(quant_c_file, std::cout);
    std::cout << "Creating parser\n";
    TranslationUnitNode ast;
    Parser test_parser(test_scanner, ast);
    //test_parser.set_debug_level(true);
    int parse_result = test_parser.parse();

    // Check program semantics and generate code if succesfully parsed
    if (parse_result == 0 && ast.isSemanticsCorrect()) {
        std::cout << "Parsing successfully completed.\n";

        // Generate LLVM IR for code
        std::cout << "Generating executable code.\n";
        ast.setName(std::string(original_path.stem()));
        ast.codegen();

        // Pull IR from Module into string stream, then write to file
        std::string IR_file_name = tmp + "/" +
                                       file_name +
                                       ".ll";
        std::string output_content;
        llvm::raw_string_ostream instruction_stream(output_content);
        instruction_stream << *(ASTNode::LLVMModule);
        instruction_stream.flush();
        std::fstream output_file(IR_file_name, std::fstream::out);
        output_file << output_content;
        output_file.close();

        // Generate an object file from the IR
        std::string obj_command = "llc -relocation-model=pic --filetype=obj ";
        obj_command += IR_file_name;
        if (system(obj_command.c_str()) != EXIT_SUCCESS) {
            std::cout << "Error generating object file";
            return -1;
        }

        // Link object file with additional code to generate executable file
        std::string obj_file_name = tmp + "/" +
                                       file_name +
                                       ".o";
        std::string link_command = "gcc -g -L${HOME}/code/C-C++/QuantC/code/build/ ";
        link_command += obj_file_name + " -lqpu_driver -lstdc++ -lcurl -ljsoncpp";
        obj_command += IR_file_name;
        if (system(link_command.c_str()) != EXIT_SUCCESS) {
            std::cout << "Error creating executable";
            return -1;
        }

        std::cout << "Code generation completed.\n";

    }
    else {
        std::cout << "Error code " << parse_result << ": Parsing ended early due to errors.\n";
    }

    // Close file
    try {
        quant_c_file.close();
    } catch (...) {
        std::cerr << "Error while closing file \"" << argv[1] << "\"\n";
        return -1;
    }

    return 0;
}
