#include <iostream>
#include <iomanip>
#include <set>
#include <algorithm>
#include <iterator>
#include <string_view>
#include <cmath>
#include <climits>
#include <exception>
#include <stdexcept>
#include <utility>
#include "symbol_table/symbol_table.hpp"
#include "ast/ast.hpp"
#include "quant_c_parse.hpp"

using namespace quant_c_ast;
using namespace symbol_table;
using namespace type_info;
using namespace quant_c_compiler;
using namespace quant_c_token;

/*
static std::pair<bool, std::string>
    isTypeCompatabile(Type expectedType, Type providedType,
                      std::string& reason) {
}
*/

// abstract superclass; only subclasses are actually created
std::shared_ptr<NullExprNode> ExprNode::NULL_VAL = std::make_shared<NullExprNode>();

ExprNode::ExprNode(ASTNodeKind nodeKind) : ASTNode::ASTNode(nodeKind) {
    type = Type();
    kind = Kind();
}

ExprNode::ExprNode(ASTNodeKind nodeKind, int line, int col) : ASTNode::ASTNode(nodeKind, line, col) {
    type = Type();
    kind = Kind();
}

ExprNode::ExprNode(ASTNodeKind nodeKind, int line, int col, Type type, Kind kind) : ASTNode::ASTNode(nodeKind, line, col) {
    this->type = type;
    this->kind = kind;
}

bool ExprNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_ExprNode &&
           nodeKind <= ASTNodeKind::AST_AssignmentOperatorNode;
} // classof
// abstract class ExprNode


NullExprNode::NullExprNode() : ExprNode::ExprNode(ASTNodeKind::AST_NullExprNode) {}
bool NullExprNode::isNull() {return true;}
void NullExprNode::unparse(int indent) { /* no action */ }
bool NullExprNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_NullExprNode;
} // classof
// class NullExprNode


BinaryOpNode::BinaryOpNode(std::shared_ptr<ExprNode> leftOperand, int operatorCode, std::shared_ptr<ExprNode> rightOperand, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_BinaryOpNode, line, col) {
    this->operatorCode = operatorCode;
    this->leftOperand = leftOperand;
    this->rightOperand = rightOperand;
}

void BinaryOpNode::printOp(int op) {
    switch (op) {
        case Parser::token_type::OR_OP:
            std::cout << " || ";
            break;
        case Parser::token_type::AND_OP:
            std::cout << " && ";
            break;
        case '<':
            std::cout << " < ";
            break;
        case '>':
            std::cout << " > ";
            break;
        case Parser::token_type::LE_OP:
            std::cout << " <= ";
            break;
        case Parser::token_type::GE_OP:
            std::cout << " >= ";
            break;
        case Parser::token_type::EQ_OP:
            std::cout << " == ";
            break;
        case Parser::token_type::NE_OP:
            std::cout << " != ";
            break;
        case '+':
            std::cout << " + ";
            break;
        case '-':
            std::cout << " - ";
            break;
        case '*':
            std::cout << " * ";
            break;
        case '/':
            std::cout << " / ";
            break;
        case '%':
            std::cout << " % ";
            break;
        case ',':
            std::cout << " , ";
            break;
        case '&':
            std::cout << " & ";
            break;
        case '|':
            std::cout << " | ";
            break;
        case '^':
            std::cout << " ^ ";
            break;
        case Parser::token_type::LEFT_OP:
            std::cout << " << ";
            break;
        case Parser::token_type::RIGHT_OP:
            std::cout << " >> ";
            break;
        case Parser::token_type::SWAP:
            std::cout << " <> ";
            break;
        case Parser::token_type::ISWAP:
            std::cout << " <I> ";
            break;
        case Parser::token_type::PSWAP:
            std::cout << " <P> ";
            break;
        default:
            throw std::invalid_argument("printOp: case not found");
    }
}

void BinaryOpNode::unparse(int indent) {
    std::cout << "(";
    leftOperand->unparse();
    printOp(operatorCode);
    rightOperand->unparse();
    std::cout << ")";
}
bool BinaryOpNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_BinaryOpNode;
} // classof

void BinaryOpNode::checkSemantics() {

    // Check operand semantics
    leftOperand->checkSemantics();
    rightOperand->checkSemantics();
    /*
    // Check to make sure arrays aren't used
    std::string arrayMessage = "Cannot use arrays as operands with logical operators.";
    mustNotBe(leftOperand->type.isArray,
                error() + arrayMessage);
    mustNotBe(rightOperand->type.isArray,
                error() + arrayMessage);

    // Check to make sure sub-expression(s) are not uninitialized variables
    mustNotBe((leftOperand->kind.value == KindValue::UNINITVAR) || (rightOperand->kind.value == KindValue::UNINITVAR),
               error() + "Cannot use uninitialized variable(s) within an operator expression.");
    */

    // Check to make sure sub-expression(s) are not labels
    bool isLeftLabel = leftOperand->kind.value == KindValue::LABEL;
    bool isRightLabel = rightOperand->kind.value == KindValue::LABEL;
    mustNotBe(isLeftLabel || isRightLabel,
               error() + "Cannot use label(s) within an operator expression.");

    // Check to make sure that quantum and classical values aren't used together
    bool isLeftQuantum = leftOperand->type.storage ==
                         StorageSpecVal::QUANTUM;
    bool isRightQuantum = rightOperand->type.storage ==
                         StorageSpecVal::QUANTUM;
    mustBe(!(isLeftQuantum ^ isRightQuantum), error() + "Quantum and classical values cannot be used in the same expression");

    switch (operatorCode) {
        case Parser::token_type::OR_OP:
        case Parser::token_type::AND_OP: {
            // Set expression type and kind
            type = Type(TypeValue::BOOLEAN);
            kind = Kind(KindValue::VALUE);
            //std::string message = "Operands of logical operations must be boolean expressions.";

            // Check for valid operand types
            //if (!isLeftLabel) {
            //    typeMustBe(leftOperand->type.value, TypeValue::BOOLEAN,
            //                error() + message);
            //}
            //if (!isRightLabel) {
            //    typeMustBe(rightOperand->type.value, TypeValue::BOOLEAN,
            //                error() + message);
            //}

            break;
        }
        case '<':
        case '>':
        case Parser::token_type::LE_OP:
        case Parser::token_type::GE_OP:
        case Parser::token_type::EQ_OP:
        case Parser::token_type::NE_OP: {
            // Set expression type and kind
            type = Type(TypeValue::BOOLEAN);
            kind = Kind(KindValue::VALUE);

            /*
            std::string messageType = "Operands of relational operations must be type integer, character, or boolean.";
            std::string messageSame = "Operands of relational operations must either be both boolean type or arithmetic types (integer and character).";
            std::vector<TypeValue> typeList = {TypeValue::INTEGER, TypeValue::CHARACTER, TypeValue::BOOLEAN};
            std::vector<TypeValue> arithmeticTypeList = {TypeValue::INTEGER, TypeValue::CHARACTER};

            // Check for valid operand types
            if (!isLeftLabel) {
            typeMustBe(leftOperand->type.value, typeList,
                        error() + messageType);
            }
            if (!isRightLabel) {
            typeMustBe(rightOperand->type.value, typeList,
                        error() + messageType);
            }

            // Check for corresponding operand types (either both bool or both int/char)
            // if both operands have a valid type

            if ((std::find(typeList.begin(), typeList.end(),
                          leftOperand->type.value) != typeList.end()) &&
                (std::find(typeList.begin(), typeList.end(),
                          rightOperand->type.value) != typeList.end())) {
                if (leftOperand->type.value == TypeValue::BOOLEAN) {
                    typesMustBeEqual(leftOperand->type.value, rightOperand->type.value,
                                error() + messageSame);
                }
                else {
                    typesMustBeSimilar(leftOperand->type.value, rightOperand->type.value, arithmeticTypeList,
                                error() + messageSame);
                }
            }
            */

            break;
        }
        case '+':
        case '-':
        case '*':
        case '/':
        case '%': {

            // Set expression type and kind
            type = Type(TypeValue::INTEGER);
            kind = Kind(KindValue::VALUE);

            /*
            std::string message = "Operands of arithmetic operations must be type integer or character.";
            std::vector<TypeValue> typeList = {TypeValue::INTEGER, TypeValue::CHARACTER};

            // Check for valid operand types
            if (!isLeftLabel) {
            typeMustBe(leftOperand->type.value, typeList,
                        error() + message);
            }
            if (!isRightLabel) {
            typeMustBe(rightOperand->type.value, typeList,
                        error() + message);
            }
            */
            break;
        }

        case '&':
        case '|':
        case '^': {
            if (isLeftQuantum && isRightQuantum) {
                // TODO: handle size information, final type, and final sign?
                type = Type(TypeValue::CUSTOM, true, TypeQualityVal::NONE,
                            StorageSpecVal::QUANTUM);
                kind = Kind(KindValue::VALUE);
            }
            else {
                // Assume result is classical by default
                // TODO: handle sign info and storage type(?)
                type = Type(TypeValue::BOOLEAN);
                kind = Kind(KindValue::VALUE);
            }
            break;
        }

        case Parser::token_type::LEFT_OP:
        case Parser::token_type::RIGHT_OP: {
            type = Type(TypeValue::INTEGER);
            kind = Kind(KindValue::VALUE);
            // TODO: Add semantics checks for shifting
            break;
        }

        default: {
            mustBe(false, error() + "Invalid operator.");
            type = Type(TypeValue::ERROR);
        }

    }

} // checkSemantics

llvm::Value* BinaryOpNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    llvm::Value *binaryExprCode = nullptr,
                 *leftOperandVal = leftOperand->codegen(),
                 *rightOperandVal = rightOperand->codegen();

    // Left operand type info
    bool LHSFloat = leftOperand->type.value == TypeValue::FLOAT ||
                    leftOperand->type.value == TypeValue::DOUBLE;
    bool LHSSigned = leftOperand->type.isSigned;

    // Right operand type info
    bool RHSFloat = rightOperand->type.value == TypeValue::FLOAT ||
                    rightOperand->type.value == TypeValue::DOUBLE;
    bool RHSSigned = rightOperand->type.isSigned;

    switch (operatorCode) {
        case Parser::token_type::OR_OP: {
            llvm::Value* LHSFalseConst = Builder.getIntN(
                        leftOperandVal->getType()->getPrimitiveSizeInBits(),
                        0);
            llvm::Value* RHSFalseConst = Builder.getIntN(
                        rightOperandVal->getType()->getPrimitiveSizeInBits(),
                        0);
            llvm::Value* LHSBool = Builder.CreateICmpNE(leftOperandVal,
                                                        LHSFalseConst);
            llvm::Value* RHSBool = Builder.CreateICmpNE(rightOperandVal,
                                                        RHSFalseConst);
            binaryExprCode = Builder.CreateOr(LHSBool, RHSBool);
            break;
        }
        case '|': {
            binaryExprCode = Builder.CreateOr(leftOperandVal, rightOperandVal);
            break;
        }
        case Parser::token_type::AND_OP: {
            llvm::Value* LHSFalseConst = Builder.getIntN(
                        leftOperandVal->getType()->getPrimitiveSizeInBits(),
                        0);
            llvm::Value* RHSFalseConst = Builder.getIntN(
                        rightOperandVal->getType()->getPrimitiveSizeInBits(),
                        0);
            llvm::Value* LHSBool = Builder.CreateICmpNE(leftOperandVal,
                                                        LHSFalseConst);
            llvm::Value* RHSBool = Builder.CreateICmpNE(rightOperandVal,
                                                        RHSFalseConst);
            binaryExprCode = Builder.CreateAnd(LHSBool, RHSBool);
            break;
        }
        case '&': {
            binaryExprCode = Builder.CreateAnd(leftOperandVal, rightOperandVal);
            break;
        }
        case '<': {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFCmpULT(leftOperandVal, rightOperandVal);
            }
            else if (LHSSigned || RHSSigned) {
                binaryExprCode = Builder.CreateICmpSLT(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateICmpULT(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case '>': {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFCmpUGT(leftOperandVal, rightOperandVal);
            }
            else if (LHSSigned || RHSSigned) {
                binaryExprCode = Builder.CreateICmpSGT(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateICmpUGT(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case Parser::token_type::LE_OP: {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFCmpULE(leftOperandVal, rightOperandVal);
            }
            else if (LHSSigned || RHSSigned) {
                binaryExprCode = Builder.CreateICmpSLE(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateICmpULE(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case Parser::token_type::GE_OP: {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFCmpUGE(leftOperandVal, rightOperandVal);
            }
            else if (LHSSigned || RHSSigned) {
                binaryExprCode = Builder.CreateICmpSGE(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateICmpUGE(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case Parser::token_type::EQ_OP: {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFCmpUEQ(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateICmpEQ(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case Parser::token_type::NE_OP: {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFCmpUNE(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateICmpNE(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case '+': {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFAdd(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateAdd(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case '-': {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFSub(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateSub(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case '*': {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFMul(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateMul(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case '/': {
            if ( LHSFloat || RHSFloat) {
                binaryExprCode = Builder.CreateFDiv(leftOperandVal, rightOperandVal);
            }
            else if ( LHSSigned || RHSSigned ) {
                binaryExprCode = Builder.CreateSDiv(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateUDiv(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case '%': {
            if ( LHSSigned || RHSSigned ) {
                binaryExprCode = Builder.CreateSRem(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateURem(leftOperandVal, rightOperandVal);
            }
            break;
        }
        case ',': {
            Builder.Insert(llvm::cast<llvm::Instruction>(leftOperandVal));
            binaryExprCode = Builder.Insert(llvm::cast<llvm::Instruction>(rightOperandVal));
            break;
        }
        case '^': {
            binaryExprCode = Builder.CreateXor(leftOperandVal, rightOperandVal);
            break;
        }
        case Parser::token_type::LEFT_OP: {
            binaryExprCode = Builder.CreateShl(leftOperandVal, rightOperandVal);
            break;
        }
        case Parser::token_type::RIGHT_OP: {
            if (LHSSigned) {
                binaryExprCode = Builder.CreateAShr(leftOperandVal, rightOperandVal);
            }
            else {
                binaryExprCode = Builder.CreateAShr(leftOperandVal, rightOperandVal);
            }
            break;
        }
        default:
            break;
    }

    return binaryExprCode;
} // codegen

QuantumValue BinaryOpNode::quantum_codegen(QuantumContext& context) {
    auto leftQubits = leftOperand->quantum_codegen(context);
    auto rightQubits = rightOperand->quantum_codegen(context);
    switch (operatorCode) {
        case '^': {
            // TODO: check for if classical or quantum
            return context.buildBinaryGateOp(QuantumOpCode::CNOT,
                                      leftQubits,
                                      rightQubits);
            break;
        }
        default: break;
    }
    return QUANTUM_NULL_VAL;
}
// class BinaryOpNode

UnaryOpNode::UnaryOpNode(std::shared_ptr<OperatorNode> op, std::shared_ptr<ExprNode> operand, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_UnaryOpNode, line, col) {
    this->opObject = op;
    this->operand = operand;
}

void UnaryOpNode::unparse(int indent) {
    std::cout << "( ";
    opObject->unparse();
    std::cout << " ";
    operand->unparse();
    std::cout << " )";
}
bool UnaryOpNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_UnaryOpNode;
} // classof

void UnaryOpNode::checkSemantics() {
    operand->checkSemantics();

    /*
    // Check to make sure sub-expression are not uninitialized variables
    mustNotBe(operand->kind.value == KindValue::UNINITVAR,
               error() + "Cannot use an uninitialized variable within an operator expression.");


    // Check to make sure sub-expression(s) are not uninitialized variables
    mustNotBe(operand->kind.value == KindValue::LABEL,
               error() + "Cannot use a label within an operator expression.");

    // Check to make sure arrays aren't used
    mustNotBe(operand->type.isArray,
              error() + "Cannot use arrays as operands with logical operators.");
    */
    auto opObjectPtr = opObject.get();
    if(llvm::isa<UnaryOperatorNode>(opObjectPtr)) {
        int opCode = llvm::cast<UnaryOperatorNode>(opObjectPtr)->opCode;
        switch(opCode) {
            case '!': {
                // Set expression type and kind
                type = Type(TypeValue::BOOLEAN);
                kind = Kind(KindValue::VALUE);
            }
            default: {
                type = Type(TypeValue::INTEGER);
                kind = Kind(KindValue::VALUE);
            }
        }
    }
    else if (llvm::isa<QuantumUnaryOperatorNode>(opObjectPtr)) {
        // Get type information based on type of operand and
        // how the operation itself affects storage
        opObject->checkSemantics();
        type = Type(opObject->type.value, operand->type.isSigned,
                    TypeQualityVal::NONE, opObject->type.storage);
        kind = Kind(KindValue::VALUE);
    }
    else {
        mustBe(false, error() + "Invalid operator.");
        type = Type(TypeValue::ERROR);
    }
} // checkSemantics

llvm::Value* UnaryOpNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    llvm::Value *operandVal = operand->codegen();
    llvm::Value *operandAddr = nullptr;
    llvm::Value* unaryOpCode = nullptr;

    // If the expression is an identifier, load its address as well
    auto operandPtr = operand.get();
    if(llvm::isa<IdentNode>(operandPtr)) {
        IdentNode* idNode = llvm::cast<IdentNode>(operandPtr);
        idNode->genAddr = true;
        operandAddr = operand->codegen();
    }

    // Operand type info
    bool isFloat = operand->type.value == TypeValue::FLOAT ||
                    operand->type.value == TypeValue::DOUBLE;
    // bool isSigned = operand->type.isSigned;

    auto opObjectPtr = opObject.get();
    if(llvm::isa<UnaryOperatorNode>(opObjectPtr)) {
        UnaryOperatorNode* opNode = llvm::cast<UnaryOperatorNode>(opObjectPtr);
        int opCode = opNode->opCode;
        bool isPostfix = opNode->isPostfix;

        switch(opCode) {
            case '&': break; // TODO: Implement ops for addresses
            case '*': break;
            case '+': unaryOpCode = operandVal;
            case '-': {
                if (isFloat) {
                    unaryOpCode = Builder.CreateFNeg(operandVal);
                }
                else {
                    unaryOpCode = Builder.CreateNeg(operandVal);
                }
                break;
            }
            case '~': {
                unaryOpCode = Builder.CreateNot(operandVal);
                break;
            }
            case '!': {
                unaryOpCode = Builder.CreateICmpEQ(
                                operandVal,
                                Builder.getInt32(0));
                break;
            }
            case Parser::token_type::INC_OP: {
                llvm::Value* incVal = Builder.CreateAdd(
                                        Builder.getInt32(1),
                                        operandVal);
                Builder.CreateStore(incVal, operandAddr);
                unaryOpCode = (isPostfix) ? operandVal : incVal;
                break;
            }
            case Parser::token_type::DEC_OP: {
                llvm::Value* incVal = Builder.CreateSub(
                                        Builder.getInt32(1),
                                        operandVal);
                Builder.CreateStore(incVal, operandAddr);
                unaryOpCode = (isPostfix) ? operandVal : incVal;
                break;
            }
        }
    }

    return unaryOpCode;
} // codegen

QuantumValue UnaryOpNode::quantum_codegen(QuantumContext& context) {
    // TODO: Handle "normal" unary ops for quantum machine (if possible)
    // and multiple qubit values
    typedef QuantumUnaryOperatorNode::QuantumUnaryOp QuantumUnaryOp;

    auto opObjectPtr = opObject.get();
    if (llvm::isa<QuantumUnaryOperatorNode>(opObjectPtr)) {
        bool isMeasure = false;
        // TODO: handle this when measuring into classical memory
        // for now, just read into default return value bit
        QuantumUnaryOperatorNode* quant_op = llvm::cast<QuantumUnaryOperatorNode>(opObjectPtr);
        if(quant_op->opCode == QuantumUnaryOp::MEASURE) {
            isMeasure = true;
        }

        // Generate code for operand and op (including mods applied)
        auto operandVal = operand->quantum_codegen(context);
        if (isMeasure) {
            return context.buildMeasurement(operandVal);
        }
        else {
            quant_op->quantum_codegen(context);

            // Map unary op to gate op code
            QuantumOpCode gateOp = QuantumOpCode::I;
            switch(quant_op->opCode) {
                case QuantumUnaryOp::IDENTITY : gateOp = QuantumOpCode::I; break;
                case QuantumUnaryOp::X_GATE   : gateOp = QuantumOpCode::X; break;
                case QuantumUnaryOp::Y_GATE   : gateOp = QuantumOpCode::Y; break;
                case QuantumUnaryOp::Z_GATE   : gateOp = QuantumOpCode::Z; break;
                case QuantumUnaryOp::HADAMARD : gateOp = QuantumOpCode::H; break;
                case QuantumUnaryOp::PHASE    : gateOp = QuantumOpCode::PHASE; break;
                case QuantumUnaryOp::S_GATE   : gateOp = QuantumOpCode::S; break;
                case QuantumUnaryOp::T_GATE   : gateOp = QuantumOpCode::T; break;
                case QuantumUnaryOp::RX       : gateOp = QuantumOpCode::RX; break;
                case QuantumUnaryOp::RY       : gateOp = QuantumOpCode::RY; break;
                case QuantumUnaryOp::RZ       : gateOp = QuantumOpCode::RZ; break;
                default: break;
            }
            return context.buildUnaryGateOp(gateOp, operandVal);
        }
    }
    return QUANTUM_NULL_VAL;
}

// class UnaryOpNode

ConditionalExprNode::ConditionalExprNode(std::shared_ptr<ExprNode> condition,
                std::shared_ptr<ExprNode> lExpr,
                std::shared_ptr<ExprNode> rExpr,
                int line, int col) :
                ExprNode(AST_ConditionalExprNode, line, col) {
    this->condition = condition;
    this->lExpr = lExpr;
    this->rExpr = rExpr;
}
void ConditionalExprNode::unparse(int indent) {
    std::cout << " ( ";
    condition->unparse();
    std::cout << " ) ? ";
    lExpr->unparse();
    std::cout << " : ";
    rExpr->unparse();
}
void ConditionalExprNode::checkSemantics() {
    // TODO: check expression types are compatible
    type = lExpr->type;
    kind = Kind(KindValue::VALUE);
}

llvm::Value* ConditionalExprNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    llvm::Value* conditionVal = condition->codegen();
    llvm::Value* falseConst = Builder.getIntN(
                            conditionVal->getType()->getPrimitiveSizeInBits(),
                            0);
    conditionVal = Builder.CreateICmpNE(conditionVal, falseConst);

    // get current function to associate other statements with it
    llvm::Function* parentFunction = Builder.GetInsertBlock()->getParent();

    // Create blocks for true/false expressions
    llvm::BasicBlock* trueBlock = llvm::BasicBlock::Create(LLVMContext, "conditional.true", parentFunction);
    llvm::BasicBlock* falseBlock = llvm::BasicBlock::Create(LLVMContext, "conditional.false");
    llvm::BasicBlock* mergeBlock = llvm::BasicBlock::Create(LLVMContext, "conditional.merge");
    Builder.CreateCondBr(conditionVal, trueBlock, falseBlock);

    // Generate code for true/false expressions
    Builder.SetInsertPoint(trueBlock);
    auto lExprVal = lExpr->codegen();
    Builder.CreateBr(mergeBlock);
    trueBlock = Builder.GetInsertBlock(); // Update true block

    parentFunction->getBasicBlockList().push_back(falseBlock);
    Builder.SetInsertPoint(falseBlock);
    auto rExprVal = rExpr->codegen();
    Builder.CreateBr(mergeBlock);
    falseBlock = Builder.GetInsertBlock(); // Update false block

    // Generate merge point and value for conditional expression
    parentFunction->getBasicBlockList().push_back(mergeBlock);
    Builder.SetInsertPoint(mergeBlock);
    llvm::PHINode *mergeVal = Builder.CreatePHI(lExprVal->getType(), 2, "conditional.value");
    mergeVal->addIncoming(lExprVal, trueBlock);
    mergeVal->addIncoming(rExprVal, falseBlock);

    return mergeVal;
}

// class ConditionalExprNode


AsgExprNode::AsgExprNode(std::shared_ptr<ExprNode> lvalue, int operatorCode, std::shared_ptr<ExprNode> rvalue, int line, int col): ExprNode::ExprNode(ASTNodeKind::AST_AsgExprNode, line, col) {
    this->lvalue = lvalue;
    this->rvalue = rvalue;
    this->assignOp = operatorCode;
}

void AsgExprNode::printOp(int op) {
    switch(op) {
        case '=':                               std::cout << '='; break;
        case Parser::token_type::MUL_ASSIGN:    std::cout << "*="; break;
        case Parser::token_type::DIV_ASSIGN:    std::cout << "/="; break;
        case Parser::token_type::MOD_ASSIGN:    std::cout << "%="; break;
        case Parser::token_type::ADD_ASSIGN:    std::cout << "+="; break;
        case Parser::token_type::SUB_ASSIGN:    std::cout << "-="; break;
        case Parser::token_type::LEFT_ASSIGN:   std::cout << "<<="; break;
        case Parser::token_type::RIGHT_ASSIGN:  std::cout << ">>="; break;
        case Parser::token_type::AND_ASSIGN:    std::cout << "&="; break;
        case Parser::token_type::XOR_ASSIGN:    std::cout << "^="; break;
        case Parser::token_type::OR_ASSIGN:     std::cout << "|="; break;
    }
}
void AsgExprNode::unparse(int indent) {
    lvalue->unparse();
    std::cout << " ";
    printOp(assignOp);
    std::cout << " ";
    rvalue->unparse();
}

void AsgExprNode::checkSemantics() {
    // Check sub-expression semantics
    lvalue->isLvalue = true;
    lvalue->checkSemantics();
    rvalue->checkSemantics();
    auto lvaluePtr = lvalue.get();
    if (llvm::isa<IdentNode>(lvaluePtr)) {
        IdentNode* ident = llvm::cast<IdentNode>(lvaluePtr);
        ident->genAddr = true;
    }

    // Check that left side is a variable
    mustBe( (lvalue->kind.value == KindValue::UNINITVAR) ||
            (lvalue->kind.value == KindValue::INITVAR),
            error() + "expression is not assignable");

    // TODO: check type compatability between left and right side

    // Set type to be type of left side
    type = lvalue->type;
    kind = lvalue->kind;
}

llvm::Value* AsgExprNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    llvm::Value *lvalueCode = lvalue->codegen(),
                 *rvalueCode = rvalue->codegen(),
                 *assignCode = nullptr;

    bool isVolatile = (lvalue->type.quality == TypeQualityVal::VOLATILE ||
                       lvalue->type.quality == TypeQualityVal::VOLATILE_CONST);
    bool isSigned = lvalue->type.isSigned;
    auto lvalueType = lvalueCode->getType();
    bool isFloat = lvalueType->isFloatTy() || lvalueType->isDoubleTy();

    // Load the value of the LHS expression if needed
    // (not simply assign rvalue to lvalue)
    llvm::Value* lvalueLoad = nullptr;
    if (assignOp != '=') {
        if(llvm::isa<llvm::AllocaInst>(lvalueCode)) {
            llvm::AllocaInst* lallocVal =
                                    llvm::cast<llvm::AllocaInst>(lvalueCode);
            lvalueLoad = Builder.CreateLoad(lallocVal->getAllocatedType(),
                                            lallocVal,
                                            isVolatile);
        }
    }

    switch(assignOp) {
        case '=': {
            assignCode = rvalueCode;
            break;
        }
        case Parser::token_type::MUL_ASSIGN: {
            if (isFloat) {
                assignCode = Builder.CreateFMul(lvalueLoad, rvalueCode);
            }
            else {
                assignCode = Builder.CreateMul(lvalueLoad, rvalueCode);
            }
            break;
        }
        case Parser::token_type::DIV_ASSIGN: {
            if (isFloat) {
                assignCode = Builder.CreateFDiv(lvalueLoad, rvalueCode);
            }
            else if (isSigned) {
                assignCode = Builder.CreateSDiv(lvalueLoad, rvalueCode);
            }
            else {
                assignCode = Builder.CreateUDiv(lvalueLoad, rvalueCode);
            }
            break;
        }
        case Parser::token_type::MOD_ASSIGN: {
            if (isSigned) {
                assignCode = Builder.CreateSRem(lvalueLoad, rvalueCode);
            }
            else {
                assignCode = Builder.CreateURem(lvalueLoad, rvalueCode);
            }
            break;
        }
        case Parser::token_type::ADD_ASSIGN: {
            if (isFloat) {
                assignCode = Builder.CreateFAdd(lvalueLoad, rvalueCode);
            }
            else {
                assignCode = Builder.CreateAdd(lvalueLoad, rvalueCode);
            }
            break;
        }
        case Parser::token_type::SUB_ASSIGN: {
            if (isFloat) {
                assignCode = Builder.CreateFSub(lvalueLoad, rvalueCode);
            }
            else {
                assignCode = Builder.CreateSub(lvalueLoad, rvalueCode);
            }
            break;
        }
        case Parser::token_type::LEFT_ASSIGN: {
            assignCode = Builder.CreateShl(lvalueLoad, rvalueCode);
            break;
        }
        case Parser::token_type::RIGHT_ASSIGN: {
            if (isSigned) {
                assignCode = Builder.CreateAShr(lvalueLoad, rvalueCode);
            }
            else {
                assignCode = Builder.CreateAShr(lvalueLoad, rvalueCode);
            }
            break;
        }
        case Parser::token_type::AND_ASSIGN: {
            assignCode = Builder.CreateAnd(lvalueLoad, rvalueCode);
            break;
        }
        case Parser::token_type::XOR_ASSIGN: {
            assignCode = Builder.CreateXor(lvalueLoad, rvalueCode);
            break;
        }
        case Parser::token_type::OR_ASSIGN: {
            assignCode = Builder.CreateOr(lvalueLoad, rvalueCode);
            break;
        }
        default: break;
    }

    // Store value to asssign to lvalue
    Builder.CreateStore(assignCode, lvalueCode, isVolatile);

    // Return expression stored into lvalue
    return assignCode;

}

quant_c_ast::QuantumValue AsgExprNode::quantum_codegen(QuantumContext& context) {
    auto lvalueVal = lvalue->quantum_codegen(context);
    auto rvalueVal = rvalue->quantum_codegen(context);

    if (lvalueVal.isClassical && rvalueVal.isClassical) {
        context.buildMove(lvalueVal, rvalueVal);
        return lvalueVal;
    }
    // TODO: double check if there are other valid cases;
    // Could allow quantum <- classical
    return QUANTUM_NULL_VAL;
}

bool AsgExprNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_AsgExprNode;
}

// class AsgExprNode


CastNode::CastNode(std::shared_ptr<TypeNameNode> resultType, std::shared_ptr<ExprNode> operand, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_CastNode, line, col) {
    this->resultType = resultType;
    this->operand = operand;
}

void CastNode::unparse(int indent) {
    resultType->unparse();
    std::cout << " ( ";
    operand->unparse();
    std::cout << " ) ";
}
bool CastNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_CastNode;
} // classof
void CastNode::checkSemantics() {
    // TODO: add check for type compatability
    type = resultType->generateType();
    kind = operand->kind;
}

llvm::Value* CastNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    auto exprVal = operand->codegen();
    auto resultTypeGen = type.codegen_type(LLVMContext);

    if (resultTypeGen->isPointerTy() && llvm::isa<llvm::AllocaInst>(exprVal)) {
        // Create space for pointer
        auto newPtr = Builder.CreateAlloca(resultTypeGen);

        // Create instructions to create pointer to expression and
        // store it the new allocated space
        /*
        std::vector<llvm::Value*> indexRef;
        indexRef.push_back(Builder.getInt32(0));
        auto pointerToVal = Builder.CreateGEP(exprVal->getType(), exprVal, indexRef);
        */
        Builder.CreateStore(exprVal, newPtr);
        return newPtr;
    } else {
        // Error
        return nullptr;
    }
}
// class CastNode

CallExprNode::CallExprNode(std::shared_ptr<ExprNode> funcExpr, std::vector<std::shared_ptr<ExprNode>> funcArgs, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_CallExprNode, line, col, Type(TypeValue::UNKNOWN), Kind(KindValue::FUNC)) {
    this->funcExpr = funcExpr;
    this->funcArgs = funcArgs;
}

void CallExprNode::unparse(int indent) {
    funcExpr->unparse();
    std::cout << " ( ";

    for (auto arg : funcArgs) {
        arg->unparse();
        std::cout << " , ";
    }

    // Back up to delete the last comma if arguments are printed
    if (funcArgs.size() > 0) {
        std::cout << "\b\b\b";
    }
    std::cout << " ) ";
}
bool CallExprNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_CallExprNode;
} // classof
void CallExprNode::checkSemantics() {
    // Check the validity of the function expression
    // and set the type/kind of this expression
    funcExpr->checkSemantics();
    Type funcExprType = funcExpr->type;
    type = funcExprType.subType; // Return type of the function
    if(funcExprType.storage == StorageSpecVal::QUANTUM) {
        type.storage = StorageSpecVal::QUANTUM;
    }
    else {
        type.storage = StorageSpecVal::AUTO;
    }
    kind = Kind(KindValue::VALUE);

    // If the function expression is an identifier, verify that it is in scope
    auto exprPtr = funcExpr.get();
    if(llvm::isa<IdentNode>(exprPtr)) {
        IdentNode* funcNameNode = llvm::cast<IdentNode>(exprPtr);
        try {
            semanticTable.globalLookup(funcNameNode->idname);
        }
        catch (FailedLookupException &lookupEx) {
            mustBe(false, error() + "Function called is not declared.");
            auto errorType = std::make_shared<Type>(TypeValue::ERROR);
            type = Type(errorType, std::vector<Type>());
        }
    }

    // Check if the resulting expression is a function
    // Add name to error message if the expression is an identifier
    // TODO: modify to include function pointers
    std::string name = "";
    if(llvm::isa<IdentNode>(exprPtr)) {
        IdentNode* funcNameNode = llvm::cast<IdentNode>(exprPtr);
        name = "'" + funcNameNode->idname + "'";
    }
    mustBe(funcExprType.isFunction, error() + "called object " + name +
                                    " is not a function or function pointer");

    // Check that the correct number of arguments are given
    int actualArgsSize = funcArgs.size();
    int formalArgsSize = funcExprType.functionArgTypes.size();
    bool tooFewArgs = actualArgsSize < formalArgsSize;
    bool tooManyArgs = actualArgsSize > formalArgsSize;

    if (tooFewArgs) {
        mustNotBe(tooFewArgs, error() + "Too few arguments to function");
    }
    if (tooManyArgs && !funcExprType.isVarArg) {
        mustNotBe(tooFewArgs, error() + "Too many arguments to function");
    }

    if (!tooFewArgs && (!tooManyArgs || funcExprType.isVarArg)) {
        int argcount = 1;
        auto expressionIter = funcArgs.begin();
        auto typeIter = funcExprType.functionArgTypes.begin();

        while (expressionIter != funcArgs.end()) {
            // Obtain information about the argument
            auto expr = *expressionIter;
            expr->checkSemantics();
            expressionIter = std::next(expressionIter, 1);
            if (typeIter != funcExprType.functionArgTypes.end()) {
                auto expectedType = *typeIter;
                typeIter = std::next(typeIter, 1);

                // Check types match between formal and actual arguments
                // TODO: loosen type restrictions
                Type givenType = expr->type;
                mustBe((givenType.get_base_type_value() ==
                        expectedType.get_base_type_value()) &&
                       (givenType.isArray == expectedType.isArray) &&
                       (givenType.isFunction == expectedType.isFunction),
                       error() + "incompatable type for argument " +
                                 std::to_string(argcount) +
                                 " of function call");
            }
            argcount++;
        }
    }
    /*
    // Check if name has been previously defined in higher scope
    SymbolInfo id;
    try {
        id = semanticTable.globalLookup(funcExpr->idname);

        // Update name and call expression with current info
        funcExpr->checkSemantics();
        type = id.type;
        kind = Kind(KindValue::VALUE);

        // Make sure the name in this call refers to a function
        bool isFunc = (id.kind.value == KindValue::FUNC) && id.type.isFunction;
        mustBe(isFunc, error() + "The identifier " + funcExpr->idname + " does not refer to a function.");

        if (isFunc) {
            // Make sure the function does have a valid return type to be valid as an expression
            std::vector<TypeValue> typeList = {TypeValue::INTEGER, TypeValue::CHARACTER, TypeValue::BOOLEAN};
            typeMustBe(id.type.value, typeList, error() + "Functions called in an expression must have a return type.");

            // Check that the correct number of arguments are given
            int actualArgsSize = funcArgs.size();
            int formalArgsSize = id.type.functionArgTypes.size();
            bool sameSize = actualArgsSize == formalArgsSize;
            mustBe(sameSize, error() + "Function " + funcExpr->idname + " called with "
                             + std::to_string(actualArgsSize) + " actual parameter(s), but expects "
                             + std::to_string(formalArgsSize) + ".\n\tFunction signature: " + std::string(id.type));
            if (sameSize) {
                int argcount = 0;
                auto expressionIter = funcArgs.begin();
                auto typeIter = id.type.functionArgTypes.begin();

                while (expressionIter != funcArgs.end() &&
                       typeIter != id.type.functionArgTypes.end()) {
                    auto expr = *expressionIter;
                    auto next_type = *typeIter;
                    expressionIter = std::next(expressionIter, 1);
                    typeIter = std::next(typeIter, 1);

                    // Check that argument is not an uninitialized variable
                    mustNotBe(expr->kind.value == KindValue::UNINITVAR,
                       error() + "Argument " + std::to_string(argcount) + " of " + funcExpr.idname +
                       " is uninitialized.");

                    // Check types match between formal and actual arguments
                    mustBe((expr->type.value == next_type.value) && (expr->type.isArray == next_type.isArray), error() +
                        "Argument " + std::to_string(argcount) + " of " + funcExpr.idname + " was expected to be of type " +
                        std::string(next_type) + " but was provided an expression of type " +
                        std::string(expr->type) + ".\n\tFunction signature: " + std::string(id.type));
                    argcount++;
                }
            }
        }
    }
    catch (...) {
        // Lookup failed; mark as an error
        mustBe(false, error() + "Function called is not defined.");
        funcExpr->type = Type(TypeValue::ERROR);
    }
    */
}
llvm::Value* CallExprNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    // Generate values for arguments
    std::vector<llvm::Value *> arg_vals;
    for (auto& arg : funcArgs) {
        llvm::Value* argVal;
        auto argPtr = arg.get();
        if (llvm::isa<IdentNode>(argPtr) && arg->type.isArray) {
            auto identNode = llvm::cast<IdentNode>(argPtr);
            identNode->genAddr = true;
            argVal = arg->codegen();
            std::vector<llvm::Value*> startIndex;
            startIndex.push_back(Builder.getInt32(0));
            startIndex.push_back(Builder.getInt32(0));
            argVal = Builder.CreateGEP(argVal->getType()->getContainedType(0), argVal, startIndex);
        }
        else {
            argVal = arg->codegen();
        }
        arg_vals.push_back(argVal);
    }


    auto funcExprPtr = funcExpr.get();
    if(llvm::isa<IdentNode>(funcExprPtr)) {
        // Get identifer node and determine whether this is a call to
        // a quantum function or classical function
        IdentNode* funcNameNode = llvm::cast<IdentNode>(funcExprPtr);
        if(funcNameNode->id_info.type.storage == StorageSpecVal::QUANTUM) {
            // Insert quantum function handler call with string as the progam
            // TODO: handle arguments
            auto symbolTable = LLVMModule->getValueSymbolTable();
            llvm::Function* funcName =
                llvm::cast<llvm::Function>(
                    symbolTable.lookup("quant_com"));
            // Find quantc string
            std::string quantum_var_name = std::string(".str.quantc");
            llvm::Value* quantum_instructions = llvm::cast<llvm::Value>(LLVMModule->getNamedGlobal(
                    quantum_var_name)->getInitializer());
            // build array ref for arguments to driver function call
            arg_vals.push_back(
                Builder.CreateGlobalStringPtr(funcNameNode->idname, ".str"));
            arg_vals.push_back(quantum_instructions);
            llvm::ArrayRef<llvm::Value *> arg_vals_array(arg_vals);

            return Builder.CreateCall(funcName->getFunctionType(),
                               funcName,
                               arg_vals_array);
        }
        else {
            // Convert vector into LLVM object
            llvm::ArrayRef<llvm::Value *> arg_vals_array(arg_vals);
            // Lookup the function value in LLVM module and
            // Generate call to that
            auto symbolTable = LLVMModule->getValueSymbolTable();
            llvm::Function* funcName =
                llvm::cast<llvm::Function>(
                    symbolTable.lookup(funcNameNode->idname));
            return Builder.CreateCall(funcName->getFunctionType(),
                               funcName,
                               arg_vals_array);
        }

    }
    else {
        // TODO: handle case where expression simply refers to some sort of
        // function pointer
        return nullptr;
    }
}

quant_c_ast::QuantumValue CallExprNode::quantum_codegen(QuantumContext& context) {
    std::vector<QuantumValue> arg_vals;
    for (auto& arg : funcArgs) {
        arg_vals.push_back(arg->quantum_codegen(context));
    }

    auto funcExprPtr = funcExpr.get();
    if(llvm::isa<IdentNode>(funcExprPtr)) {
        IdentNode* funcNameNode = llvm::cast<IdentNode>(funcExprPtr);
        std::string funcName = funcNameNode->idname;
        context.buildCircuitCall(funcName, arg_vals);
    }
    return QUANTUM_NULL_VAL;
}
// class CallExprNode

InitializerListNode::InitializerListNode(
    std::vector<std::shared_ptr<ExprNode>> initalizerList,
    int line, int col) :
ExprNode(ASTNodeKind::AST_InitializerListNode, line, col,
         Type(TypeValue::CUSTOM), Kind(KindValue::VALUE)) {
    this->initalizerList = initalizerList;
}

void InitializerListNode::unparse(int indent) {
    std::cout << " { ";
    for (auto& expr : initalizerList) {
        expr->unparse();
        std::cout << " , ";
    }
    // Back up output and print out final text
    if(initalizerList.size() > 0) {
        std::cout << "\b\b\b";
    }
    std::cout << " } ";
}

void InitializerListNode::checkSemantics() {
    Type firstType;
    for (auto& expr : initalizerList) {
        expr->checkSemantics();
    }
}

llvm::Value* InitializerListNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    if(isStruct) {
        // Do nothing; not implemented yet
    }
    else {
        std::vector<llvm::Constant*> elementVec;
        llvm::Type* elementType = nullptr;
        for (auto& element : initalizerList) {
            llvm::Constant* elementVal =
                    llvm::cast<llvm::Constant>(element->codegen());
            elementVec.push_back(elementVal);
            if (elementType == nullptr) {
                elementType = elementVal->getType();
            }
        }

        llvm::ArrayType* arrayConstType = llvm::ArrayType::get(elementType,
                                                        initalizerList.size());
        // NOTE: probably need to make a global value since a constant
        llvm::Constant* constArray = llvm::ConstantArray::get(arrayConstType, elementVec);
        llvm::GlobalVariable* arrayGlobal =
                                    new llvm::GlobalVariable(*LLVMModule,
                                                             arrayConstType,
                                                             true,
                                            llvm::GlobalValue::PrivateLinkage,
                                                             constArray,
                                                             ".array");
        return arrayGlobal;
    }
    return nullptr; // Error
}

QuantumValue InitializerListNode::quantum_codegen(QuantumContext& context) {
    // TODO: fix for general case
    // Will need to create temporary register for values and
    // let buildQuantumInit handle moving values to each qubit
    auto listVal = QuantumValue(0, true);
    Type bitVecSize = Type(std::make_shared<Type>(TypeValue::BOOLEAN),
                           initalizerList.size());
    return listVal;
}

bool InitializerListNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_InitializerListNode;
}

// class InitializerListNode


std::shared_ptr<NullIdentNode> IdentNode::NULL_VAL = std::make_shared<NullIdentNode>();
IdentNode::IdentNode() : ExprNode::ExprNode(ASTNodeKind::AST_IdentNode) {
    idname = "";
}

IdentNode::IdentNode(std::string idname, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_IdentNode, line, col, Type(TypeValue::UNKNOWN), Kind(KindValue::UNINITVAR)) {
    this->idname = idname;
}

void IdentNode::unparse(int indent) {
    std::cout << idname;
}
bool IdentNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_IdentNode;
} // classof

void IdentNode::checkSemantics() {
    try {
        SymbolInfo& id = semanticTable.globalLookup(idname);

        type = id.type;
        kind = id.kind;
        id_info = id; // Saves symbol table entry
    } catch(...) {
        mustBe(false, error() + idname + " undeclared.");
        type = Type(TypeValue::ERROR);
        kind = Kind();
        id_info = SymbolInfo();
    }
} // checkSemantics

// TODO: modify all code to make this return the lookup value
llvm::Value* IdentNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    // Get the value of the variable with that name
    SymbolValue symValue = codegenTable.globalLookup(idname);
    auto variableVal = symValue.codegenValue;

    // TODO: may need to account for Arguments by doing special load
    // And change codegenTable to loaded value
    // Determine whether to return the variable's address or value

    if (llvm::isa<llvm::Argument>(variableVal)) {
        auto argVal = llvm::cast<llvm::Argument>(variableVal);
        bool isVolatile = ((type.quality == TypeQualityVal::VOLATILE) ||
                           (type.quality == TypeQualityVal::VOLATILE_CONST));

        // Need to store value to location allocated in function body
        // Use that temporary value for future uses of the argument
        auto tempVal = Builder.CreateAlloca(argVal->getType());
        Builder.CreateStore(argVal, tempVal, isVolatile);
        codegenTable.insert(SymbolValue(argVal->getName(), tempVal));
        variableVal = tempVal;
    }

    if (genAddr) {
        return variableVal;
    }
    else if (llvm::isa<llvm::AllocaInst>(variableVal)) {
        auto allocaInst = llvm::cast<llvm::AllocaInst>(variableVal);
        bool isVolatile = ((type.quality == TypeQualityVal::VOLATILE) ||
                           (type.quality == TypeQualityVal::VOLATILE_CONST));

        return Builder.CreateLoad(allocaInst->getAllocatedType(),
                                      allocaInst, isVolatile, idname);
    }

    return nullptr; // Error

} // codegen

QuantumValue IdentNode::quantum_codegen(QuantumContext& context) {
    return context.getQuantumValue(id_info.get_name());
} // quantum_codegen
// class IdentNode


NullIdentNode::NullIdentNode() { /* no action */ }
bool NullIdentNode::isNull() {
    return true;
}
void NullIdentNode::unparse(int indent) { /* no action */ }
bool NullIdentNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_NullIdentNode;
} // classof
// class NullIdentNode

IndexExprNode::IndexExprNode() : ExprNode::ExprNode(ASTNodeKind::AST_IndexExprNode) {
    this->varName = ExprNode::NULL_VAL;
    this->subscriptVal = ExprNode::NULL_VAL;
}

IndexExprNode::IndexExprNode(std::shared_ptr<ExprNode> varName, std::shared_ptr<ExprNode> subscriptVal, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_IndexExprNode, line, col) {
    this->varName = varName;
    this->subscriptVal = subscriptVal;
}

IndexExprNode::IndexExprNode(std::shared_ptr<ExprNode> varName, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_IndexExprNode, line, col) {
    this->varName = varName;
    this->subscriptVal =  std::shared_ptr<ExprNode>(ExprNode::NULL_VAL);
}

void IndexExprNode::unparse(int indent) {
    varName->unparse();
    std::cout << " [ ";
    subscriptVal->unparse();
    std::cout << " ] ";
}
bool IndexExprNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_IndexExprNode;
} // classof

void IndexExprNode::checkSemantics() {
    varName->isLvalue = isLvalue;
    varName->checkSemantics();
    auto varNamePtr = varName.get();
    if(llvm::isa<IdentNode>(varNamePtr)) {
        IdentNode* nameNode = llvm::cast<IdentNode>(varNamePtr);
        nameNode->genAddr = true;
    }
    else if(llvm::isa<IndexExprNode>(varNamePtr)) {
        IndexExprNode* indexNode = llvm::cast<IndexExprNode>(varNamePtr);
        indexNode->genAddr = true;
    }
    subscriptVal->checkSemantics();
    type = *(varName->type.subType);
    if (isLvalue) {
        kind = KindValue::INITVAR;
    }
    else {
        kind = KindValue::VALUE;
    }
    /*
    // Check subscript if present
    if (!subscriptVal->isNull()) {

        // Check if subscript is of appropriate type
        std::vector<TypeValue> typeList = {TypeValue::INTEGER, TypeValue::CHARACTER};
        typeMustBe(subscriptVal->type.value, typeList, error() + "Only integer and character expressions can be used as index values.");

        // Check that subscript is not an array
        mustNotBe(subscriptVal->type.isArray, error() + "Cannot use arrays as an index; the index expression must be a single integer or character value." );

        // Check if subscript is used on an array
        if (varName->type.value != TypeValue::ERROR) {
            mustBe(varName->type.isArray, error() + "Cannot subscript an identifier that isn't an array.");
        }

        // Make sure that the type shown by this node reflects the fact that it is
        // referring to an element of an array
        type = Type(varName->type.value);

    }
    */

} // checkSemantics

llvm::Value* IndexExprNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    llvm::Value* ptrVal = varName->codegen();
    llvm::Value* indexVal = subscriptVal->codegen();

    llvm::Type* elementType = ptrVal->getType()->getContainedType(0);
    std::vector<llvm::Value*> indexVec;
    indexVec.push_back(Builder.getInt32(0));
    indexVec.push_back(indexVal);
    llvm::Value* ptrToElement = Builder.CreateInBoundsGEP(elementType,
                                                  ptrVal, indexVec);
    bool isVolatile = (varName->type.quality == TypeQualityVal::VOLATILE ||
                       varName->type.quality == TypeQualityVal::VOLATILE_CONST);

    // Return address or value depending on
    // if it used to store a value or
    // if this expression is meant to refer to a value vs a sub-array
    if (isLvalue || genAddr) {
        return ptrToElement;
    }
    else {
        return Builder.CreateLoad(ptrToElement->getType()->getContainedType(0),
                                  ptrToElement, isVolatile);
    }

}

quant_c_ast::QuantumValue IndexExprNode::quantum_codegen(QuantumContext& context) {
    auto ptrVal = varName->quantum_codegen(context);
    auto indexVal = subscriptVal->quantum_codegen(context);
    if (!(ptrVal.isClassical) && indexVal.isIntLiteral) {
        // Index qubitAddresses sequences and
        // reduce it down to that single index
        auto elementLocation = ptrVal.qubitAddressSequence[indexVal.intVal];
        std::vector<std::string> indexLoc;
        indexLoc.push_back(elementLocation);
        ptrVal.qubitAddressSequence = indexLoc;
        return ptrVal;
    }
    return QUANTUM_NULL_VAL;
}

// TODO: implement quantum_codegen for Index
// class IndexExprNode

StrLitNode::StrLitNode(std::string strval, std::string reducedStrval, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_StrLitNode, line, col, Type(TypeValue::CHARACTER, (int)reducedStrval.length()), Kind(KindValue::VALUE)) {
    this->strval = strval;
    this->reducedStrval = reducedStrval;
}

void StrLitNode::unparse(int indent) {
    std::cout << strval;
}
bool StrLitNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_StrLitNode;
} // classof

llvm::Value* StrLitNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    return Builder.CreateGlobalStringPtr(reducedStrval, ".str");
}
// class StrLitNode

IntLitNode::IntLitNode() : ExprNode::ExprNode(ASTNodeKind::AST_IntLitNode) {
    this->intval = 0;
    this->size = quant_c_token::IntLitSize::INT;
    this->isSigned = true;
}

IntLitNode::IntLitNode(std::intmax_t intval, quant_c_token::IntLitSize size, bool isSigned, int line, int col) :
ExprNode::ExprNode(ASTNodeKind::AST_IntLitNode, line, col, Type(TypeValue::INTEGER), Kind(KindValue::VALUE)) {
    this->intval = intval;
    this->size = size;
    this->isSigned = isSigned;
}

void IntLitNode::unparse(int indent) {
    std::cout << std::to_string(intval);
}

bool IntLitNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_IntLitNode;
} // classof

llvm::Value* IntLitNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    const unsigned int byte_width = CHAR_BIT;
    unsigned int bit_width;
    switch(size) {
        case IntLitSize::INT:
            bit_width = sizeof(int)*byte_width;
            break;
        case IntLitSize::LONG:
            bit_width = sizeof(long)*byte_width;
            break;
        case IntLitSize::LONGLONG:
            bit_width = sizeof(long long)*byte_width;
            break;
    }

    return llvm::ConstantInt::get(LLVMContext,
                                  llvm::APInt(bit_width, intval, isSigned));
}

quant_c_ast::QuantumValue IntLitNode::quantum_codegen(QuantumContext& context) {
    return QuantumValue(intval, isSigned);
}
// class IntLitNode

FloatLitNode::FloatLitNode() : ExprNode::ExprNode(ASTNodeKind::AST_FloatLitNode) {
    this->floatValue = 0;
    this->size = quant_c_token::FloatLitSize::FLOAT;
}

FloatLitNode::FloatLitNode(long double val, quant_c_token::FloatLitSize size, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_FloatLitNode, line, col, Type(TypeValue::FLOAT), Kind(KindValue::VALUE)) {
    this->floatValue = val;
    this->size = size;
}

void FloatLitNode::unparse(int indent) {
    //TODO: add unparse code
}
bool FloatLitNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_FloatLitNode;
} // classof
// class FloatLitNode


CharLitNode::CharLitNode(char charval, int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_CharLitNode, line, col, Type(TypeValue::CHARACTER), Kind(KindValue::VALUE)) {
    this->charval = charval;
}

std::string CharLitNode::printCharValue() {
    switch(charval) {
        case '\n': return "'\\n'";
        case '\t': return "'\\t'";
        case '\\': return "'\\\\'";
        case '\'': return "'\\\''";
        default: return std::string("'") + charval + "'";
    }
}

void CharLitNode::unparse(int indent) {
    std::cout << printCharValue();
}
bool CharLitNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_CharLitNode;
} // classof
// class CharLitNode

TrueNode::TrueNode(int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_TrueNode, line, col, Type(TypeValue::BOOLEAN), Kind(KindValue::VALUE)) {}

void TrueNode::unparse(int indent) {
    std::cout << "true";
}
bool TrueNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_TrueNode;
} // classof
// class TrueNode

FalseNode::FalseNode(int line, int col) : ExprNode::ExprNode(ASTNodeKind::AST_FalseNode, line, col, Type(TypeValue::BOOLEAN), Kind(KindValue::VALUE)) {}

void FalseNode::unparse(int indent) {
    std::cout << "false";
}
bool FalseNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_FalseNode;
} // classof
// class FalseNode


// TODO: implement operatorNode for constructors
OperatorNode::OperatorNode(ASTNodeKind nodeKind): ExprNode(nodeKind) {};
OperatorNode::OperatorNode(ASTNodeKind nodeKind, int line, int col) :
    ExprNode(nodeKind, line, col) {};
bool OperatorNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_OperatorNode &&
           nodeKind <= ASTNodeKind::AST_AssignmentOperatorNode;
}


UnaryOperatorNode::UnaryOperatorNode(int opCode):
OperatorNode(AST_UnaryOperatorNode) {
    this->opCode = opCode;
    this->isPostfix = false;
}
UnaryOperatorNode::UnaryOperatorNode(int opCode, bool post, int line, int col):
OperatorNode(AST_UnaryOperatorNode, line, col) {
    this->opCode = opCode;
    this->isPostfix = post;
}
void UnaryOperatorNode::unparse(int indent) {
    switch(opCode) {
        case Parser::token_type::INC_OP: std::cout << "++"; break;
        case Parser::token_type::DEC_OP: std::cout << "--"; break;
        default:
            std::cout << (char)opCode;
    }
}
bool UnaryOperatorNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_UnaryOperatorNode;
}


QuantumUnaryOperatorNode::QuantumUnaryOperatorNode(QuantumUnaryOp op, int line, int col)
: OperatorNode(ASTNodeKind::AST_QuantumUnaryOperatorNode, line, col) {
    isInverted = false;
    opCode = op;
    opMods = std::vector<std::shared_ptr<QuantumUnaryPostfixNode>>();
}

// TODO: handle unary ops with phase info
QuantumUnaryOperatorNode::QuantumUnaryOperatorNode(bool invert,
                         QuantumUnaryOp op,
                         std::vector<std::shared_ptr<QuantumUnaryPostfixNode>>
                         modifiers,
                         int line, int col)
: OperatorNode(ASTNodeKind::AST_QuantumUnaryOperatorNode, line, col) {
    isInverted = invert;
    opCode = op;
    opMods = modifiers;
}

void QuantumUnaryOperatorNode::unparse(int indent) {
    std::string op_string = "";
    op_string += (isInverted) ? "~>" : "" ;
    switch(opCode) {
        case QuantumUnaryOp::IDENTITY: std::cout << op_string + "%I"; break;
        case QuantumUnaryOp::X_GATE  : std::cout << op_string + "%X"; break;
        case QuantumUnaryOp::Y_GATE  : std::cout << op_string + "%Y"; break;
        case QuantumUnaryOp::Z_GATE  : std::cout << op_string + "%Z"; break;
        case QuantumUnaryOp::HADAMARD: std::cout << op_string + "%H"; break;
        case QuantumUnaryOp::PHASE   : std::cout << op_string + "%PHASE"; break;
        case QuantumUnaryOp::S_GATE  : std::cout << op_string + "%S"; break;
        case QuantumUnaryOp::T_GATE  : std::cout << op_string + "%T"; break;
        case QuantumUnaryOp::RX      : std::cout << op_string + "%RX"; break;
        case QuantumUnaryOp::RY      : std::cout << op_string + "%RY"; break;
        case QuantumUnaryOp::RZ      : std::cout << op_string + "%RZ"; break;
        case QuantumUnaryOp::MEASURE : std::cout << op_string + "%M"; break;
    }
    for (auto& mod : opMods) {
        mod->unparse();
    }
}

void QuantumUnaryOperatorNode::checkSemantics() {
    // Note that type value is a dummy value; only the storage type is important
    switch(opCode) {
        case QuantumUnaryOp::MEASURE  : type = Type(TypeValue::INTEGER);
                                        break;
        default                       : type = Type(TypeValue::INTEGER, true,
                                                    TypeQualityVal::NONE,
                                                    StorageSpecVal::QUANTUM);
                                        break;
    }
}

QuantumValue QuantumUnaryOperatorNode::quantum_codegen(QuantumContext& context) {

    std::vector<QuantumOpMod> codegenMods;

    // Add inversion op to gate if needed
    if (isInverted) {
        context.insertOpMod(QuantumOpMod(OpMod::INVERSE));
    }

    for (auto& mod : opMods) {
        mod->quantum_codegen(context);
    }

    return QUANTUM_NULL_VAL;
}

bool QuantumUnaryOperatorNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_QuantumUnaryOperatorNode;
} // classof
// class QuantumUnaryOperatorNode

QuantumUnaryPostfixNode::QuantumUnaryPostfixNode(IdentNode id, int line, int col) : OperatorNode(ASTNodeKind::AST_QuantumUnaryPostfixNode, line, col) {
    this->id = id;
}

void QuantumUnaryPostfixNode::unparse(int indent) {
    std::cout << "+C( " + id.idname + " )";
}
void QuantumUnaryPostfixNode::checkSemantics() {
    // TODO: check that id is quantum
}

quant_c_ast::QuantumValue QuantumUnaryPostfixNode::quantum_codegen(QuantumContext& context) {
    auto quantumVal = context.getQuantumValue(id.idname);
    context.insertOpMod(QuantumOpMod(OpMod::CONTROL, quantumVal));
    return QUANTUM_NULL_VAL;
}

bool QuantumUnaryPostfixNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_QuantumUnaryPostfixNode;
}
// class QuantumUnaryPostfixNode

// end ExprNode subclasses
