#include <iostream>
#include <iomanip>
#include <algorithm>
#include <iterator>
#include <string_view>
#include <cmath>
#include <exception>
#include <stdexcept>
#include "symbol_table/symbol_table.hpp"
#include "ast/ast.hpp"
#include "quant_c_parse.hpp"

using namespace quant_c_ast;
using namespace symbol_table;
using namespace type_info;
using namespace quant_c_compiler;

#define MAX_TYPE_SPECS 2;
#define MAX_STORAGE_SPECS 1;
#define MAX_TYPE_QUALIFIERS 2;

type_info::Type ArgDeclNode::generateArgType() {
    Type argTypeSpecs = Type();
    argTypeSpecs.isScalar = true;
    bool set_signedness = false;
    bool set_quality = false;
    bool set_storage = false;

    for (auto& spec : declSpecs) {
        auto type_node = spec.get();
        if (llvm::isa<TypeSpecNode>(type_node)) {
            if (llvm::isa<LongTypeNode>(type_node)) {
                if (argTypeSpecs.value == TypeValue::UNKNOWN ||
                    argTypeSpecs.value == TypeValue::INTEGER) {
                    argTypeSpecs.value = TypeValue::LONG;
                }
                else if (argTypeSpecs.value == TypeValue::LONG) {
                    argTypeSpecs = TypeValue::LONGLONG;
                }
                else {
                    mustBe(false, error() + "both 'long' and '" +
                           argTypeSpecs.get_value_string() +
                           "' in declaration specifiers");
                    return Type(TypeValue::ERROR);
                }
            }
            else if (llvm::isa<ShortTypeNode>(type_node)) {
                if (argTypeSpecs.value == TypeValue::UNKNOWN ||
                    argTypeSpecs.value == TypeValue::INTEGER) {
                    argTypeSpecs.value = TypeValue::SHORT;
                }
                else if (argTypeSpecs.value == TypeValue::SHORT) {
                    mustBe(false, error() + "duplicate 'short'");
                    return Type(TypeValue::ERROR);
                }
                else {
                    mustBe(false, error() + "both 'short' and '" +
                           argTypeSpecs.get_value_string() +
                           "' in declaration specifiers");
                    return Type(TypeValue::ERROR);
                }
            }
            else if (llvm::isa<SignedTypeNode>(type_node)) {
                if (set_signedness) {
                    if (argTypeSpecs.isSigned) {
                        mustBe(false, error() + "duplicate 'signed'");
                    }
                    else {
                        mustBe(false, error() + "both 'signed' and 'unsigned' in declaration specifiers");
                    }
                    return Type(TypeValue::ERROR);
                } else {
                    argTypeSpecs.isSigned = true;
                    set_signedness = true;
                }
            }
            else if (llvm::isa<UnsignedTypeNode>(type_node)) {
                if (set_signedness) {
                    if (!argTypeSpecs.isSigned) {
                        mustBe(false, error() + "duplicate 'unsigned'");
                    }
                    else {
                        mustBe(false, error() + "both 'signed' and 'unsigned' in declaration specifiers");
                    }
                    return Type(TypeValue::ERROR);
                } else {
                    argTypeSpecs.isSigned = false;
                    set_signedness = true;
                }
            }
            else if (argTypeSpecs.value == TypeValue::UNKNOWN) {
                argTypeSpecs.value = spec->type.value;
            }
            else {
                mustBe(false, error() +
                       "two or more data types in declaration specifiers");
                return Type(TypeValue::ERROR);
            }
        }
        else if (llvm::isa<StorageClassSpecNode>(type_node)) {
            auto spec = llvm::cast<StorageClassSpecNode>(type_node);
            if (!(llvm::isa<RegisterNode>(spec) || llvm::isa<QuantumNode>(spec))) {
                mustBe(false, error() +
                    "storage class specified for parameter");
            }
            else if (llvm::isa<QuantumNode>(spec)) {
                argTypeSpecs.storage = StorageSpecVal::QUANTUM;
                set_storage = true;
            }
            else {
                argTypeSpecs.storage = StorageSpecVal::REGISTER;
                set_storage = true;
            }
        }
        else if (llvm::isa<TypeQualifierNode>(type_node)) {
            set_quality = true;
            if (argTypeSpecs.quality == TypeQualityVal::VOLATILE_CONST) {
                continue; // Any additional qualifiers are ignored
            }
            else if (llvm::isa<VolatileNode>(type_node)) {
                argTypeSpecs.quality =
                    (argTypeSpecs.quality == TypeQualityVal::CONST) ?
                    TypeQualityVal::VOLATILE_CONST :
                    TypeQualityVal::VOLATILE;

            } else { // llvm::isa<ConstNode>(spec)
                argTypeSpecs.quality =
                    (argTypeSpecs.quality == TypeQualityVal::VOLATILE) ?
                    TypeQualityVal::VOLATILE_CONST :
                    TypeQualityVal::CONST;
            }
        }
        else {
            mustBe(false, error() + "Unimplemented type used");
            return Type(TypeValue::ERROR);
        }
    }

    // Establish defaults if left implicit in the type specs
    if (argTypeSpecs.value == TypeValue::UNKNOWN) {
        argTypeSpecs.value = TypeValue::INTEGER;
    }
    if (!set_signedness) {
        argTypeSpecs.isSigned = true;
    }
    if (!set_storage) {
        argTypeSpecs.storage = StorageSpecVal::REGISTER;
    }
    if (!set_quality) {
        argTypeSpecs.quality = TypeQualityVal::NONE;
    }
    return argTypeSpecs;
}
// abstract superclass; only subclasses are actually created

ArgDeclNode::ArgDeclNode(ASTNodeKind nodeKind, DeclSpecifierList declSpecs) :
ASTNode::ASTNode(nodeKind) {
    this->declSpecs = declSpecs;
}
ArgDeclNode::ArgDeclNode(ASTNodeKind nodeKind, DeclSpecifierList declSpecs, int line, int col) : ASTNode::ASTNode(nodeKind, line, col) {
    this->declSpecs = declSpecs;
}
bool ArgDeclNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_ArgDeclNode &&
           nodeKind < ASTNodeKind::AST_StmtNode;
} // classof
// class ArgDeclNode

/*
ArrayArgDeclNode::ArrayArgDeclNode(IdentNode argName, std::shared_ptr<TypeNode> elementType, int line, int col) : ArgDeclNode::ArgDeclNode(line, col, argName) {
    this->elementType = elementType;
}

void ArrayArgDeclNode::unparse(int indent) {
    argName.unparse();
    std::cout << " [ ] ";
    elementType->unparse();
}

void ArrayArgDeclNode::checkSemantics() {
    // Check if variable name has been previously defined in same scope
    SymbolInfo id;
    try {
        id = semanticTable.localLookup(argName.idname);

        // Lookup should throw an error before this.
        // If not, the variable has been previously defined; mark as an error
        mustBe(false, error() + id.get_name() + " is already declared.");
        argName.type = Type(TypeValue::ERROR);
    }
    catch (...) {

        // Pick type info for identifier name based on argument type
        Type unsizedArray = Type(elementType->type.value, 0);
        id = SymbolInfo(argName.idname, Kind(KindValue::FUNCARG), unsizedArray);

        try {
            semanticTable.insert(id);
        } catch (DuplicateException &dupEx){
            // can't happen
        } catch (EmptySTException &emptyEx){
            // can't happen
        }

        // Ensure name gets correct type information
        argName.checkSemantics();
    }
} // checkSemantics

// class ArrayArgDeclNode


ValArgDeclNode::ValArgDeclNode(IdentNode argName, std::shared_ptr<TypeNode> argType, int line, int col) : ArgDeclNode::ArgDeclNode(line, col, argName) {
    this->argType = argType;
}

void ValArgDeclNode::unparse(int indent) {
    argName.unparse();
    std::cout << " ";
    argType->unparse();
}

void ValArgDeclNode::checkSemantics() {
    // Check if variable name has been previously defined in same scope
    SymbolInfo id;
    try {
        id = semanticTable.localLookup(argName.idname);

        // Lookup should throw an error before this.
        // If not, the variable has been previously defined; mark as an error
        mustBe(false, error() + id.get_name() + " is already declared.");
        argName.type = Type(TypeValue::ERROR);
    }
    catch (...) {

        // Pick type info for identifier name based on argument type
        id = SymbolInfo(argName.idname, Kind(KindValue::FUNCARG), argType->type);

        try {
            semanticTable.insert(id);
        } catch (DuplicateException &dupEx){
            // can't happen
        } catch (EmptySTException &emptyEx){
            // can't happen
        }

        // Ensure name gets correct type information
        argName.checkSemantics();
    }
}
// class ValArgDeclNode

*/

NamedArgDeclNode::NamedArgDeclNode(DeclSpecifierList declSpecs,
                                  std::shared_ptr<DeclaratorNode> decl,
                                  int line, int col)
: ArgDeclNode(ASTNodeKind::AST_NamedArgDeclNode, declSpecs, line, col) {
    this->decl = decl;
}

void NamedArgDeclNode::unparse(int indent) {
    for (auto& specs : declSpecs) {
        specs->unparse();
        std::cout << " ";
    }
    decl->unparse();
}

bool NamedArgDeclNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_NamedArgDeclNode;
} // classof

void NamedArgDeclNode::checkSemantics() {
    // Check declarator is valid and fetch information
    decl->checkSemantics();
    try {
        SymbolInfo& id = semanticTable.globalLookup(decl->id_name);

        // Generate and update arg type in symbol table
        // based on preliminary type found by declarator check
        auto argBaseType = generateArgType();
        /*
        if (id.type.isArray) {
            id.type = Type(argTypePtr, id.type.size,
                           argTypePtr->quality, argTypePtr->storage);

        }
        else if (id.type.isFunction) {
            id.type = Type(argTypePtr, id.type.functionArgTypes,
                           id.type.isVarArg,
                           argTypePtr->quality, argTypePtr->storage);
        }
        else if (id.type.isPointer) {
            id.type = Type(argTypePtr);
        }
        else {
            id.type = argTypePtr;
        }
        */
        id.type.set_base_type(argBaseType);

        // Set id and type information
        id_info = id;
        argType = id.type;
    } catch (FailedLookupException &lookupEx) {
        // Shouldn't occur
        mustBe(false, "Unexpected failure while checking program; this must be a compiler bug.");
    }
}

/*
llvm::Type* NamedArgDeclNode::get_type() {

}
*/
// class NamedArgDeclNode


AbstractArgDeclNode::AbstractArgDeclNode(DeclSpecifierList declSpecs,
                                  std::shared_ptr<AbstractDeclaratorNode> decl,
                                  int line, int col) :
ArgDeclNode(ASTNodeKind::AST_AbstractArgDeclNode, declSpecs, line, col) {
    this->decl = decl;
}

void AbstractArgDeclNode::unparse(int indent) {
    for (auto& specs : declSpecs) {
        specs->unparse();
        std::cout << " ";
    }
    // decl->unparse();
}

bool AbstractArgDeclNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_AbstractArgDeclNode;
} // classof

void AbstractArgDeclNode::checkSemantics() {
    argType = generateArgType();
    // TODO: implement AbstractDeclNode and type info
}
// class AbstractArgDeclNode


TypeArgDeclNode::TypeArgDeclNode(DeclSpecifierList declSpecs,
                                 int line, int col) :
ArgDeclNode(ASTNodeKind::AST_TypeArgDeclNode, declSpecs, line, col) {}

void TypeArgDeclNode::unparse(int indent) {
    for (auto& specs : declSpecs) {
        specs->unparse();
        std::cout << " ";
    }
}

bool TypeArgDeclNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_TypeArgDeclNode;
} // classof

void TypeArgDeclNode::checkSemantics() {
    // TODO: implement checkSemantics
    argType = generateArgType();
}
// class TypeArgDeclNode

// end ArgDeclNode subclasses
