#include <iostream>
#include <iomanip>
#include <algorithm>
#include <iterator>
#include <string_view>
#include <cmath>
#include <exception>
#include <stdexcept>
#include "symbol_table/symbol_table.hpp"
#include "ast/ast.hpp"
#include "quant_c_parse.hpp"

using namespace quant_c_ast;
using namespace symbol_table;
using namespace type_info;
using namespace quant_c_compiler;

// abstract superclass; only subclasses are actually created

StmtNode::StmtNode(ASTNodeKind nodeKind) : ASTNode::ASTNode(nodeKind) {}
StmtNode::StmtNode(ASTNodeKind nodeKind, int line, int col) : ASTNode::ASTNode(nodeKind, line, col) {}
bool StmtNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_StmtNode &&
           nodeKind < ASTNodeKind::AST_ExprNode;
} // classof
std::shared_ptr<NullStmtNode> StmtNode::NULL_VAL = std::make_shared<NullStmtNode>();
// class StmtNode

NullStmtNode::NullStmtNode() : StmtNode(ASTNodeKind::AST_NullStmtNode) {}
bool NullStmtNode::isNull() { return true; }
bool NullStmtNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_NullStmtNode;
} // classof
//class NullStmtNode

/*
AsgNode::AsgNode(NameNode target, std::shared_ptr<ExprNode> source, int line, int col) : StmtNode(ASTNodeKind::AST_AsgNode, line, col) {
    this->target = target;
    this->source = source;
}


void AsgNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    target.unparse();
    std::cout << " = ";
    source->unparse();
    std::cout << " ;" << "\n";
}
bool AsgNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_AsgNode;
} // classof

void AsgNode::checkSemantics() {
    // TODO: Implement semantics checks
    // Check semantics of both parts of assign statement
    target.checkSemantics();
    source->checkSemantics();

    // If left-hand side is an invalid identifier, that's an error
    std::vector<KindValue> kindArray = {KindValue::CONST, KindValue::LABEL, KindValue::FUNC, KindValue::OTHER};
    bool isVariable = (std::find(kindArray.begin(), kindArray.end(), target.kind.value) == kindArray.end());
    mustBe(isVariable, error() + std::string("Assignment to a ") +
                       std::string(target.kind) + " is invalid.");

    // If target was an uninitialized variable, mark it as now initialized
    if (target.kind.value == KindValue::UNINITVAR && !target.varName.type.isArray) {
        target.kind.value = KindValue::INITVAR;
        target.varName.idinfo.kind.value = KindValue::INITVAR;
    }

    // Check if source is an uninitialized variable
    mustNotBe(source->kind.value == KindValue::UNINITVAR, error() + "Cannot assign using an uninitialized variable.");

    if (isVariable) {
        // Check both sides of assignment have the same type
        typesMustBeEqual(source->type.value, target.type.value,
            error() + "Both the left and right"
            + " hand sides of an assignment must "
            + "have the same type; the left is of type " + std::string(target.type)
            + ", while the right is of type " + std::string(source->type) + ".");
        // If one side is an array, make sure the sizes also match
        if ((target.type.isArray || source->type.isArray) &&
        target.kind.value != KindValue::FUNCARG &&
        source->kind.value != KindValue::FUNCARG) {
            mustBe(target.type.size == source->type.size, error() +
                   "Size mismatch; the left hand side expects exactly " +
                   std::to_string(target.type.size) + " element(s), but the right hand side provides " +
                   std::to_string(source->type.size) + " element(s).");
        }
    }


} // checkSemantics

// class AsgNode
*/

IfThenNode::IfThenNode(std::shared_ptr<ExprNode> condition, std::shared_ptr<StmtNode> thenPart, std::shared_ptr<StmtNode> elsePart, int line, int col) : StmtNode::StmtNode(ASTNodeKind::AST_IfThenNode, line, col) {
    this->condition = condition;
    this->thenPart = thenPart;
    this->elsePart = elsePart;
}


void IfThenNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "if ";
    condition->unparse();
    std::cout << "\n";
    thenPart->unparse(indent);
    if ( !(elsePart->isNull()) ) {
        std::cout << std::left << std::setw(4) << elsePart->get_line() << ":";
        genIndent(indent);
        std::cout << "else" << "\n";
    }
    elsePart->unparse(indent);
}
bool IfThenNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_IfThenNode;
} // classof
void IfThenNode::checkSemantics() {
    condition->checkSemantics();
    thenPart->checkSemantics();
    elsePart->checkSemantics();
    /*
    // check semantics of the condition expression
    condition->checkSemantics();
    typeMustBe(condition->type.value, TypeValue::BOOLEAN,
        error() + "The control expression of an if-statement must be a boolean value.");
    mustNotBe(condition->kind.value == KindValue::UNINITVAR, error() +
                  "Cannot condition an if-statement on an uninitialized variable.");

    // check then and else bodies with their own local scopes
    thenPart->checkSemantics();
    elsePart->checkSemantics();
    */
} // checkSemantics
llvm::Value* IfThenNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    llvm::Value* conditionVal = condition->codegen();
    llvm::Value* falseConst = Builder.getIntN(
                            conditionVal->getType()->getPrimitiveSizeInBits(),
                            0);
    conditionVal = Builder.CreateICmpNE(conditionVal, falseConst);

    // get current function to associate other statements with it
    llvm::Function* parentFunction = Builder.GetInsertBlock()->getParent();

    // Create blocks for then/else statements
    llvm::BasicBlock* thenBlock = llvm::BasicBlock::Create(LLVMContext, "if.then", parentFunction);
    llvm::BasicBlock* elseBlock = llvm::BasicBlock::Create(LLVMContext, "if.else");
    llvm::BasicBlock* mergeBlock = llvm::BasicBlock::Create(LLVMContext, "if.continue");
    Builder.CreateCondBr(conditionVal, thenBlock, elseBlock);

    bool continueNeeded = false;
    // Generate code for then/else parts
    Builder.SetInsertPoint(thenBlock);
    //llvm::Value* thenValue =
    thenPart->codegen();
    thenBlock = Builder.GetInsertBlock(); // Update then block
    // Don't generate break if block already returns
    if(!llvm::isa<llvm::ReturnInst>(thenBlock->back())) {
        continueNeeded = true;
        Builder.CreateBr(mergeBlock);
    }
    thenBlock = Builder.GetInsertBlock(); // Update then block

    parentFunction->getBasicBlockList().push_back(elseBlock);
    Builder.SetInsertPoint(elseBlock);
    // llvm::Value* elseValue =
    elsePart->codegen();
    elseBlock = Builder.GetInsertBlock(); // Update else block
    // Don't generate break if block already returns
    if(!llvm::isa<llvm::ReturnInst>(elseBlock->back())) {
        continueNeeded = true;
        Builder.CreateBr(mergeBlock);
    }
    elseBlock = Builder.GetInsertBlock(); // Update else block

    // Generate merge point (if needed)
    if (continueNeeded) {
        parentFunction->getBasicBlockList().push_back(mergeBlock);
        Builder.SetInsertPoint(mergeBlock);
    }
    return nullptr;
} // codegen

// class IfThenNode


ForNode::ForNode(std::shared_ptr<StmtNode> initializer_stmt, std::shared_ptr<StmtNode> condition_stmt,
std::shared_ptr<ExprNode> update_expr, std::shared_ptr<StmtNode> body, int line, int col) : StmtNode::StmtNode(ASTNodeKind::AST_ForNode, line, col) {
    this->initializer_stmt = initializer_stmt;
    this->condition_stmt = condition_stmt;
    this->update_expr = update_expr;
    this->body = body;
}

void ForNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "for ( ";
    initializer_stmt->unparse();
    condition_stmt->unparse();
    update_expr->unparse();
    std::cout << " ) \n";
    body->unparse(indent);
}
bool ForNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ForNode;
} // classof
void ForNode::checkSemantics() {
    // TODO: Re-implement semantics checks
    initializer_stmt->checkSemantics();
    condition_stmt->checkSemantics();
    update_expr->checkSemantics();
    body->checkSemantics();
    /*
    // Check semantics of label if present
    if(!label.isNull()) {
        // Check if label has been previously defined in same or higher scope and open a scope for the for-loop body
        SymbolInfo id;
        try {
            id = semanticTable.localLookup(label.idname);

            // Lookup should throw an error before this.
            // If not, the variable has been previously defined; mark as an error
            mustBe(false, error() + id.get_name() + " is already declared and can't be used as a label.");
            label.type = Type(TypeValue::ERROR);
        }
        catch (...) {
            // Add label to scope
            id = SymbolInfo(label.idname, Kind(KindValue::LABEL), Type(TypeValue::VOID));

            try {
                semanticTable.insert(id);
            } catch (DuplicateException &dupEx) {
                // can't happen
            } catch (EmptySTException &emptyEx) {
                // can't happen
            }

            // Update label information
            label.checkSemantics();

        }
    } // if(!label.isNull())

    // Check condition is a Boolean value
    condition->checkSemantics();
    typeMustBe(condition->type.value, TypeValue::BOOLEAN,
       error() + "The control expression of a for-statement must be a boolean value.");
    mustNotBe(condition->kind.value == KindValue::UNINITVAR,
       error() + "Cannot condition a for-statement on an uninitialized variable.");

    // Check validity of body
    loopBody.checkSemantics();
    */
}

llvm::Value* ForNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    // Generate initlizing code
    //llvm::Value* initVal =
    initializer_stmt->codegen();

    // Save place in prior code and
    // get parent function to indicate where new blocks should go
    llvm::BasicBlock* previousBlock = Builder.GetInsertBlock();
    llvm::Function* parentFunc = previousBlock->getParent();

    // Create block for body and have previous blokc jump to it
    // append update expression to end (if any)
    llvm::BasicBlock* bodyBlock = llvm::BasicBlock::Create(LLVMContext, "for.body", parentFunc);
    Builder.CreateBr(bodyBlock);


    Builder.SetInsertPoint(bodyBlock);
    //llvm::PHINode* forNode = Builder.CreatePHI(initVal->getType(), 2);
    //forNode->addIncoming(initVal, previousBlock);
    body->codegen();


    //llvm::Value* updateVal =
    update_expr->codegen();

    llvm::Value* conditionVal = condition_stmt->codegen();

    // Build a block of code for the condition
    //llvm::BasicBlock* bodyEnd = Builder.GetInsertBlock();
    llvm::BasicBlock* exitBlock = llvm::BasicBlock::Create(LLVMContext, "for.exit", parentFunc);

    Builder.CreateCondBr(conditionVal, bodyBlock, exitBlock);

    Builder.SetInsertPoint(exitBlock);

    //forNode->addIncoming(updateVal, bodyEnd);

    return exitBlock;

}
// class ForNode

WhileNode::WhileNode(std::shared_ptr<ExprNode> condition, std::shared_ptr<StmtNode> loopBody, bool is_do_while, int line, int col) : StmtNode(ASTNodeKind::AST_WhileNode, line, col) {
    this->condition = condition;
    this->body = body;
    this->is_do_while = is_do_while;
}

void WhileNode::unparse(int indent) {
    // TODO: verify unparse
    std::cout << std::left << std::setw(4) << linenum << ": ";
    genIndent(indent);
    if (is_do_while) {
        std::cout << "do" << "\n";
        body->unparse(indent);
        std::cout << "while" << " ( ";
        condition->unparse();
        std::cout << " ) ;";
    } else {
        std::cout << "while" << " ( ";
        condition->unparse();
        std::cout << " ) " << "\n";
        body->unparse(indent);
    }

}
bool WhileNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_WhileNode;
} // classof
void WhileNode::checkSemantics() {
    // TODO: implement semantics check
}
// class WhileNode


BreakNode::BreakNode(int line, int col) : StmtNode::StmtNode(ASTNodeKind::AST_BreakNode, line, col) {}

void BreakNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "break ;\n";
}
bool BreakNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_BreakNode;
} // classof
void BreakNode::checkSemantics() {
    // TODO: re-implement semantics checks
    /*
    // Check if label has been previously defined in higher scope
    SymbolInfo id;
    SymbolInfo localid;

    try {
        id = semanticTable.globalLookup(label.idname);
    }
    catch (...) {
        // Global lookup failed; mark as an error
        mustBe(false, error() + "Invalid label used; the label is undefined.");
        label.type = Type(TypeValue::ERROR);
    }

    try {
        localid = semanticTable.localLookup(label.idname);

        // Lookup should throw an error before this.
        // If not, the label does refer to one that this break can reach;
        // mark as an error
        mustBe(false, error() + "Invalid label used; the break statement was not contained within the body of a for-loop with said label.");
        label.type = Type(TypeValue::ERROR);
    }
    catch (...) {
        // Check name refers to a label
        mustBe(id.kind.value == KindValue::LABEL, error() + "Specified identifier is not a valid label.");

        // Update label with current info
        label.checkSemantics();
    }
    */
}

llvm::Value* BreakNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    if (breakToLocation) {
        return Builder.CreateBr(breakToLocation);
    }
    return nullptr;
} // codegen

// class BreakNode

ContinueNode::ContinueNode(int line, int col) : StmtNode::StmtNode(ASTNodeKind::AST_ContinueNode, line, col) {}

void ContinueNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "continue ; \n";
}
bool ContinueNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ContinueNode;
} // classof
void ContinueNode::checkSemantics() {
    // TODO: re-implement semantics checks
    /*
    // Check if label has been previously defined in higher scope
    SymbolInfo id;
    SymbolInfo localid;
    try {
        id = semanticTable.globalLookup(label.idname);
    }
    catch (...) {
        // Global lookup failed; mark as ab error
        mustBe(false, error() + "Invalid label used; the label is undefined.");
        label.type = Type(TypeValue::ERROR);

    }

    try {
        localid = semanticTable.localLookup(label.idname);

        // Lookup should throw an error before this.
        // If not, the label does refer to one that this continue can reach;
        // mark as an error
        mustBe(false, error() + "Invalid label used; the continue statement was not contained within the body of a for-loop with said label.");
        label.type = Type(TypeValue::ERROR);
    }
    catch (...) {
        // Check name refers to a label
        mustBe(id.kind.value == KindValue::LABEL, error() + "Specified identifier is not a valid label.");

        // Update label with current info
        label.checkSemantics();
    }
    */
}
// class ContinueNode

ReturnNode::ReturnNode(std::shared_ptr<ExprNode> returnVal) : StmtNode(ASTNodeKind::AST_ReturnNode) {
    this->returnVal = returnVal;
    this->expectedType = TypeValue::VOID;
}

ReturnNode::ReturnNode(std::shared_ptr<ExprNode> returnVal, int line, int col) : StmtNode(ASTNodeKind::AST_ReturnNode, line, col) {
    this->returnVal = returnVal;
    this->expectedType = TypeValue::VOID;
}

void ReturnNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "return ";
    returnVal->unparse();
    std::cout << " ;\n";
}

void ReturnNode::setExpectedType(type_info::Type type) {
    expectedType = type;
}

bool ReturnNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ReturnNode;
} // classof

void ReturnNode::checkSemantics() {

    // check expression and get type of return value
    // Note: the expression could be something like a function call,
    // we need to grab the base type, not the type directly
    returnVal->checkSemantics();
    auto [returnValType, hadValidSubType] = returnVal->type.get_base_type();
    if (!hadValidSubType) {
        // Means that the type had a nullptr subtype,
        // even though it shouldn't; report error
        mustBe(false, error() + "Unexpected failure while setting type information for declaration; this must be a compiler bug.");
    }

    // Check that returned expression matches expectations
    std::string errorMessage;
    bool expectingVoid = expectedType.value == TypeValue::VOID &&
                      expectedType.isScalar;
    if (expectingVoid && !returnVal->isNull()) {
        errorMessage = "'return' with a value in function returning void.";
        mustBe(returnVal->isNull(), error() + errorMessage);
    }
    else {
        // change error message based on existence of return value
        if (returnValType->value == TypeValue::UNKNOWN) {
            errorMessage = "Expected to return a value of type " +
                           expectedType.get_value_string() +
                           ", but the return statement doesn't return anything.";
        }
        else {
            errorMessage = "Expected to return a value of type " +
                           expectedType.get_value_string() +
                           ", but the return value is of type " +
                           returnValType->get_value_string() + ".";
        }

        std::vector<TypeValue> discreteTypes =
        {
            TypeValue::CHARACTER,
            TypeValue::SHORT,
            TypeValue::INTEGER,
            TypeValue::LONG,
            TypeValue::LONGLONG,
            TypeValue::FLOAT,
            TypeValue::DOUBLE,
            TypeValue::BOOLEAN,
        };
        if (std::find(discreteTypes.begin(),
                      discreteTypes.end(),
                      expectedType.value) != discreteTypes.end()) {
            typesMustBeSimilar(*returnValType, expectedType,
                               discreteTypes, error() + errorMessage);
        }


        // Check to make sure return expression doesn't refer to an array
        mustNotBe(returnValType->isArray, error() + "Functions cannot return an array.");
    }


}

llvm::Value* ReturnNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    llvm::Value* returnValCode = returnVal->codegen();
    llvm::Type* returnType = expectedType.codegen_type(LLVMContext);
    llvm::Value* returnInstruction = nullptr;

    if (returnValCode == nullptr) {
        // Set default return based on function type
        if (returnType->isVoidTy()) {
            returnInstruction = Builder.CreateRetVoid();
        }
        else {
            auto zeroVal = llvm::Constant::getNullValue(returnType);
            returnInstruction = Builder.CreateRet(zeroVal);
        }
    }
    else if (llvm::isa<llvm::AllocaInst>(returnValCode)) {
        // Create load if value is an identifier
        llvm::AllocaInst* identPtrVal =
                                llvm::cast<llvm::AllocaInst>(returnValCode);
        bool isVolatile =
                ((returnVal->type.quality == TypeQualityVal::VOLATILE) ||
                (returnVal->type.quality == TypeQualityVal::VOLATILE_CONST));

        returnValCode = Builder.CreateLoad(identPtrVal->getAllocatedType(),
                                           identPtrVal, isVolatile,
                                           identPtrVal->getName());
    }

    // Cast value to the appropriate return type if neccesary
    if (returnValCode != nullptr && (returnType != returnValCode->getType())) {

        llvm::Value* castedValue = Builder.CreateIntCast(returnValCode,
                                                   returnType,
                                                   expectedType.isSigned);
        returnInstruction = Builder.CreateRet(castedValue);
    }
    else {
        returnInstruction = Builder.CreateRet(returnValCode);
    }
    return returnInstruction;
}

QuantumValue ReturnNode::quantum_codegen(QuantumContext& context) {
    // TODO: see if additional work is needed
    auto returnExprVal = returnVal->quantum_codegen(context);
    if (!returnExprVal.isNull) {
        context.buildReturn(returnExprVal);
    }
    return QUANTUM_NULL_VAL;
}
// class ReturnNode

std::shared_ptr<NullBlockNode> BlockNode::NULL_VAL = std::make_shared<NullBlockNode>();
BlockNode::BlockNode() : StmtNode::StmtNode(ASTNodeKind::AST_BlockNode) {
    this->decls = std::vector<std::shared_ptr<DeclNode>>();
    this->stmts = std::vector<std::shared_ptr<StmtNode>>();
    this->givenScope = false;
    this->returnStatus = returnNotNeccesary;
    this->isFuncBody = false;
    this->requiredReturnType = TypeValue::VOID;
    this->containsReturn = false;
    this->alwaysReachesReturn = false;
}

BlockNode::BlockNode(std::vector<std::shared_ptr<DeclNode>> decls, std::vector<std::shared_ptr<StmtNode>> stmts, int line, int col) : StmtNode::StmtNode(ASTNodeKind::AST_BlockNode, line, col) {
    this->decls = decls;
    this->stmts = stmts;
    this->givenScope = false;
    this->returnStatus = returnNotNeccesary;
    this->isFuncBody = false;
    this->requiredReturnType = TypeValue::VOID;
    this->containsReturn = false;
    this->alwaysReachesReturn = false;
}

void BlockNode::setGivenScope(bool value) {
    givenScope = value;
}

void BlockNode::setReturnStatus(bool status, Type type) {
    returnStatus = status;
    requiredReturnType = type;
}

void BlockNode::setReturnStatus(bool status) {
    returnStatus = status;
    requiredReturnType = Type(TypeValue::VOID);
}

void BlockNode::setIsFuncBody() {
    isFuncBody = true;
}

bool BlockNode::doesContainReturn() {
    return containsReturn;
}

bool BlockNode::doesAlwaysReachReturn() {
    return alwaysReachesReturn;
}

void BlockNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "{\n";

    for (auto& decl : decls) {
        decl->unparse(indent+1);
    }
    for (auto& stmt : stmts) {
        stmt->unparse(indent+1);
    }

    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "}" << "\n";
}

bool BlockNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_BlockNode;
} // classof

void BlockNode::checkSemantics() {
    // TODO: re-implement semantics checks
    // Open a new scope if a parent node did not already provide one
    if (!givenScope) {
        semanticTable.openScope();
    }

    for (auto& vardecl : decls) {
        vardecl->checkSemantics();
    }

    // Keep track of return statements at this level of the block
    bool hasLocalReturn = false;

    for (auto& stmt : stmts) {
        auto stmt_node = stmt.get();
        if (llvm::isa<ReturnNode>(stmt_node)) {
            containsReturn = true;
            hasLocalReturn = true;
            alwaysReachesReturn = true;
            ReturnNode* caughtNode = llvm::cast<ReturnNode>(stmt_node);
            caughtNode->setExpectedType(requiredReturnType); // Pass type requirement to return
            stmt_node->checkSemantics();
        }
        else if (llvm::isa<BreakNode>(stmt_node)) continue;
        else if (llvm::isa<BlockNode>(stmt_node)) {
            BlockNode* caughtNode = llvm::cast<BlockNode>(stmt_node);
            caughtNode->setReturnStatus(returnStatus, requiredReturnType);
        }
        else if (llvm::isa<IfThenNode>(stmt_node)) {
            IfThenNode* ifThenStmt = llvm::cast<IfThenNode>(stmt_node);
            auto thenPtr = ifThenStmt->thenPart.get();
            auto elsePtr = ifThenStmt->elsePart.get();
            bool thenIsReturn = false,
                 elseIsReturn = false,
                 thenAlwaysReachesReturn = false,
                 elseAlwaysReachesReturn = false;

            // Check for return statements and sub-blocks
            if (llvm::isa<ReturnNode>(thenPtr)) {
                containsReturn = true;
                thenIsReturn = true;
                ReturnNode* caughtNode = llvm::cast<ReturnNode>(thenPtr);
                caughtNode->setExpectedType(requiredReturnType); // Pass type requirement to return
            }
            else if (llvm::isa<BlockNode>(thenPtr)) {
                BlockNode* caughtNode = llvm::cast<BlockNode>(thenPtr);
                caughtNode->setReturnStatus(returnStatus, requiredReturnType);
            }

            if (llvm::isa<ReturnNode>(elsePtr)) {
                containsReturn = true;
                elseIsReturn = true;
                ReturnNode* caughtNode = llvm::cast<ReturnNode>(elsePtr);
                caughtNode->setExpectedType(requiredReturnType); // Pass type requirement to return
            }
            else if (llvm::isa<BlockNode>(elsePtr)) {
                BlockNode* caughtNode = llvm::cast<BlockNode>(elsePtr);
                caughtNode->setReturnStatus(returnStatus, requiredReturnType);
            }

            // Check semantics
            stmt_node->checkSemantics();

            // Check that if parts are blocks, they contains return statements
            if (llvm::isa<BlockNode>(thenPtr)) {
                BlockNode* caughtNode = llvm::cast<BlockNode>(thenPtr);
                thenAlwaysReachesReturn = caughtNode->doesAlwaysReachReturn();
                if (caughtNode->doesContainReturn() && !containsReturn) {
                    containsReturn = true;
                }
            }
            if (llvm::isa<BlockNode>(elsePtr)) {
                BlockNode* caughtNode = llvm::cast<BlockNode>(elsePtr);
                elseAlwaysReachesReturn = caughtNode->doesAlwaysReachReturn();
                if (caughtNode->doesContainReturn() && !containsReturn) {
                    containsReturn = true;
                }
            }

            // If reaching a return is not guaranteed already,
            // check if both control paths reach a return
            if ((thenIsReturn || thenAlwaysReachesReturn)
                && (elseIsReturn || elseAlwaysReachesReturn)) {
                alwaysReachesReturn = true;
            }
        }
        else if (llvm::isa<SwitchNode>(stmt_node)) {
            SwitchNode* switchStmt = llvm::cast<SwitchNode>(stmt_node);
            auto switchBody = switchStmt->body.get();
            if (llvm::isa<ReturnNode>(switchBody)) {
                containsReturn = true;
                ReturnNode* caughtNode = llvm::cast<ReturnNode>(switchBody);
                caughtNode->setExpectedType(requiredReturnType); // Pass type
            }
            else if (llvm::isa<BlockNode>(switchBody)) {
                BlockNode* caughtNode = llvm::cast<BlockNode>(switchBody);
                caughtNode->setReturnStatus(returnStatus, requiredReturnType);
                caughtNode->setGivenScope(true);
            }

            // Check semantics
            stmt_node->checkSemantics();

            if (switchStmt->allBranchesReturn) alwaysReachesReturn = true;
        }
        // TODO: add check for other statements with sub-statements/blocks
        else stmt_node->checkSemantics(); // No block in this statement; simply check semantics
    }


    if (isFuncBody) {
        if (returnStatus == returnRequired && alwaysReachesReturn) {
            // Check that function can't reach the end without a return
            mustBe(alwaysReachesReturn, error() + "The body of a function with a return type reaches end of function without return");
        }
        else if (returnStatus == returnNotNeccesary &&
                 !alwaysReachesReturn && !hasLocalReturn) {
            // Add an implicit return statement to end of function bodies
            // that don't return anything and can reach the end
            // without a return statement
            std::shared_ptr<StmtNode> null_return(new ReturnNode(std::shared_ptr<ExprNode>(ExprNode::NULL_VAL)));
            stmts.push_back(null_return);
        }
    }

    // close scope if this node needed to open one
    if (!givenScope) {
        try {
            semanticTable.closeScope();
        } catch (EmptySTException &emptyEx) {
            // can't happen
        }
    }


    /*
    // Keep track of return statements at this level of the block
    bool hasLocalReturn = false;

    // Check variable declarations and statements
    for (auto vardecl : decls) {
        vardecl->checkSemantics();
    }
    for (auto stmt : stmts) {

        // Pass on necessary info about expected return type and appropriateness
        // in this block to other sub-blocks before checking semantics
        std::string stmt_class = stmt->getClassName();
        if (stmt_class == "ReturnNode") { // Set semantic expectations for the return statment
            containsReturn = true;
            hasLocalReturn = true;
            std::shared_ptr<ReturnNode> caughtNode = std::static_pointer_cast<ReturnNode>(stmt);
            caughtNode->setExpectedType(requiredReturnType); // Pass type requirement to return
            stmt->checkSemantics();
        }
        else if (stmt_class == "IfThenNode") {
            // Pass this return information onto then and else block;
            // Make return not necessary since return could be elsewhere
            std::shared_ptr<IfThenNode> caughtNode = std::static_pointer_cast<IfThenNode>(stmt);
            caughtNode->thenPart.setReturnStatus(returnNotNeccesary, requiredReturnType);
            caughtNode->elsePart.setReturnStatus(returnNotNeccesary, requiredReturnType);

            // check semantics of if statement
            stmt->checkSemantics();

            // Check if contained return
            if (!containsReturn && (caughtNode->thenPart.doesContainReturn() || caughtNode->elsePart.doesContainReturn())) {
                containsReturn = true;
            }
        }
        else if (stmt_class == "ForNode") {
            // Pass this return information onto the loop body block;
            // Make return not necessary since return could be elsewhere
            std::shared_ptr<ForNode> caughtNode = std::static_pointer_cast<ForNode>(stmt);
            caughtNode->loopBody.setReturnStatus(returnNotNeccesary, requiredReturnType);

            // check semantics of for statement
            stmt->checkSemantics();

            // Check if contained return
            if (!containsReturn && caughtNode->loopBody.doesContainReturn()) {
                containsReturn = true;
            }
        }
        else if (stmt_class == "BlockNode") {
            // Pass this return information onto the sub-block;
            // Make return not necessary since return could be elsewhere
            std::shared_ptr<BlockNode> caughtNode = std::static_pointer_cast<BlockNode>(stmt);
            caughtNode->setReturnStatus(returnNotNeccesary, requiredReturnType);

            // check semantics of for statement
            stmt->checkSemantics();

            // Check if contained return
            if (!containsReturn && caughtNode->doesContainReturn()) {
                containsReturn = true;
            }
        }
        else stmt->checkSemantics(); // No block in this statement; simply check semantics
    }

    if (returnStatus == returnRequired) {
        // Check for presence of return statement in the body
        mustBe(containsReturn, error() + "The body of a function with a return type must contain a return statement.");
    } else if (isFuncBody && !hasLocalReturn) {
        // Add an implicit return statement to function bodies that don't
        // return anything and doesn't have one
        std::shared_ptr<StmtNode> null_return(new ReturnNode(std::shared_ptr<ExprNode>(ExprNode::NULL_VAL)));
        stmts.push_back(null_return);
    }

    // close scope if this node needed to open one
    if (!givenScope) {
        try {
            semanticTable.closeScope();
        } catch (EmptySTException &emptyEx) {
            // can't happen
        }
    }
    */
}
// class BlockNode

llvm::Value* BlockNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    llvm::BasicBlock* block_code;
    parentCode = parent;
    if (parent && llvm::isa<llvm::Function>(parent)) {
        auto function_code = llvm::cast<llvm::Function>(parent);
        std::string function_name = std::string(function_code->getName());
        block_code = llvm::BasicBlock::Create(LLVMContext,
                                              function_name + "_body",
                                              function_code);
        Builder.SetInsertPoint(block_code);
    }
    else if (parent && llvm::isa<llvm::SwitchInst>(parent)) {
        succesorBlock = block_code = llvm::BasicBlock::Create(LLVMContext,
                                              "case.exit");
    }
    else {
        // TODO: handle generic block statement
    }

    // Open new scope for symbol values in body
    codegenTable.openScope();

    // TODO: handle decls
    if (!parent || !(llvm::isa<llvm::SwitchInst>(parent))) {
        for (auto& decl : decls) {
            //llvm::Value *decl_val =
            decl->codegen(block_code);
        }
    }

    // Generate statements
    for (auto& stmt : stmts) {

        // If this block is for a switch, need to pass on some information
        // about the end of the block further down before continuing codegen
        if (parent && llvm::isa<llvm::SwitchInst>(parent)) {
            auto stmtPtr = stmt.get();
            if (llvm::isa<CaseStmtNode>(stmtPtr)) {
                auto caseStmt = llvm::cast<CaseStmtNode>(stmtPtr);
                auto caseBody = caseStmt->stmt.get();
                if (llvm::isa<BlockNode>(caseBody)) {
                    auto caseBlock = llvm::cast<BlockNode>(caseBody);
                    caseBlock->succesorBlock = block_code;
                }
                else if (llvm::isa<BreakNode>(caseBody)) {
                    auto breakStmt = llvm::cast<BreakNode>(caseBody);
                    breakStmt->breakToLocation = block_code;
                }
            }
            //llvm::Value* stmt_val =
            stmt->codegen(parent);
        }

        // General case
        else {
            //llvm::Value* stmt_val =
            stmt->codegen(block_code);
        }

        // No codegen past return
        auto stmtPtr = stmt.get();
        if (llvm::isa<ReturnNode>(stmtPtr)) break;
    }

    if (parent && llvm::isa<llvm::SwitchInst>(parent)) {
        // Link exit block back to parent and set builder to there
        // if this is for a switch statement
        //if (block_code->hasNPredecessorsOrMore(1) ) {
            auto funcBody = Builder.GetInsertBlock()->getParent();
            block_code->insertInto(funcBody);
            Builder.SetInsertPoint(block_code);
        //}
    }

    // Close scope for symbol values in body
    codegenTable.closeScope();

    return block_code;
}

QuantumValue BlockNode::quantum_codegen(QuantumContext& context) {

    context.qubitAddresses.openScope();
    for (auto& decl : decls) {
        decl->quantum_codegen(context);
    }
    for (auto& stmt : stmts) {
        auto result_val = stmt->quantum_codegen(context);
    }
    context.qubitAddresses.closeScope();

    return QUANTUM_NULL_VAL;
}

NullBlockNode::NullBlockNode() { /* no action */ }
bool NullBlockNode::isNull() {return true;}
void NullBlockNode::unparse(int indent) { /* no action */ }
//class NullBlockNode

ExprStmtNode::ExprStmtNode(std::shared_ptr<ExprNode> expression, int line, int col) : StmtNode(ASTNodeKind::AST_ExprStmtNode, line, col) {
    this->expression = expression;
}

void ExprStmtNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    expression->unparse(indent);
    std::cout << " ;\n";
}
bool ExprStmtNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ExprStmtNode;
} // classof
void ExprStmtNode::checkSemantics() {
    expression->checkSemantics();
}
llvm::Value* ExprStmtNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    return expression->codegen();
}

quant_c_ast::QuantumValue ExprStmtNode::quantum_codegen(QuantumContext& context) {
    return expression->quantum_codegen(context);
}
// class ExprStmtNode


LabelStmtNode::LabelStmtNode(IdentNode label, std::shared_ptr<StmtNode> stmt, int line, int col) : StmtNode(ASTNodeKind::AST_LabelStmtNode, line, col) {
    this->label = label;
    this->stmt = stmt;
}

void LabelStmtNode::unparse(int indent) {
    // TODO: verify unparse
    std::cout << std::left << std::setw(4) << linenum << ": ";
    genIndent(indent);
    label.unparse();
    std::cout << "\n";
    stmt->unparse(indent);
}
bool LabelStmtNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_LabelStmtNode;
} // classof
void LabelStmtNode::checkSemantics() {
    // TODO: implement semantics check
}
//class LabelStmtNode

SwitchNode::SwitchNode(std::shared_ptr<ExprNode> condition, std::shared_ptr<StmtNode> body, int line, int col) : StmtNode(ASTNodeKind::AST_SwitchNode, line, col) {
    this->condition = condition;
    this->body = body;
}

void SwitchNode::unparse(int indent) {
    // TODO: verify unparse
    std::cout << std::left << std::setw(4) << linenum << ": ";
    genIndent(indent);
    std::cout << "switch" << " ( ";
    condition->unparse();
    std::cout << " ) " << "\n";
    body->unparse(indent);
}
bool SwitchNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_SwitchNode;
} // classof

void SwitchNode::checkSemantics() {
    condition->checkSemantics();

    auto stmtPtr = body.get();

    // Propagate return info to cases
    // Previous block will have already passed this onto switch prior
    if (llvm::isa<BlockNode>(stmtPtr)) {
        auto blockPtr = llvm::cast<BlockNode>(stmtPtr);
        for (auto& stmt : blockPtr->stmts) {
            auto stmtPtr = stmt.get();
            if (llvm::isa<CaseStmtNode>(stmtPtr)) {
                auto caseStmt = llvm::cast<CaseStmtNode>(stmtPtr);
                caseStmt->setReturnStatus(blockPtr->returnStatus,
                                           blockPtr->requiredReturnType);

               // Check if case is default as well
               if (caseStmt->is_default || caseStmt->leadsToDefault)
                    explicitDefault = true;
            }
            else {
                // TODO: figure out how to handle semantics for
                // general C switch syntax
                mustBe(false, "Non-case/default statement inside the body of switch statement");
            }



        }
    }

    body->checkSemantics();

    // Check to see if control flow can reach past the end of the switch
    if (llvm::isa<ReturnNode>(stmtPtr)) {
        allBranchesReturn = true;
    }
    else if (llvm::isa<BlockNode>(stmtPtr)) {
        bool throughPathExists = false;
        auto blockPtr = llvm::cast<BlockNode>(stmtPtr);
        for (auto& stmt : blockPtr->stmts) {
            auto stmtPtr = stmt.get();
            if (llvm::isa<CaseStmtNode>(stmtPtr)) {
                auto caseStmt = llvm::cast<CaseStmtNode>(stmtPtr);
                throughPathExists = !(caseStmt->alwaysReturns);
            }
            else {
                // Note: Assuming any non-case/default statements let
                // control through just in case
                throughPathExists = true;
            }

            // Found a through path; no need to check further
            if (throughPathExists) break;
        }
        allBranchesReturn = !throughPathExists;
    }
    else {
        allBranchesReturn = true;
    }



}

llvm::Value* SwitchNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    llvm::Value* switchExprVal = condition->codegen();
    int numCasesEstimate = 1;

    // Check if body is actually a block and estimate
    // the number of cases in the block
    auto bodyPtr = body.get();
    bool isBlock = llvm::isa<BlockNode>(bodyPtr);
    if (isBlock) {
        auto blockNode = llvm::cast<BlockNode>(bodyPtr);
        numCasesEstimate = blockNode->stmts.size();
    }

    llvm::SwitchInst* switchVal = Builder.CreateSwitch(switchExprVal,
                                                  nullptr,
                                                  numCasesEstimate);
    body->codegen(switchVal);

    // Set default case to block after instruction if
    // default case not specified
    // Note: this should always be a new block; switch instructions terminate
    //       blocks
    if (!explicitDefault) {
        auto exitBLock = Builder.GetInsertBlock();
        switchVal->setDefaultDest(exitBLock);
    }

    return switchVal;
} // codegen

// class SwitchNode


CaseStmtNode::CaseStmtNode(std::shared_ptr<ExprNode> case_val, std::shared_ptr<StmtNode> stmt, int line, int col) : StmtNode(ASTNodeKind::AST_CaseStmtNode, line, col) {
    this->case_val = case_val;
    this->stmt = stmt;
    this->is_default = false;
}

CaseStmtNode::CaseStmtNode(std::shared_ptr<StmtNode> stmt, int line, int col) : StmtNode(ASTNodeKind::AST_CaseStmtNode, line, col) {
    this->case_val = ExprNode::NULL_VAL;
    this->stmt = stmt;
    this->is_default = true;
}

void CaseStmtNode::unparse(int indent) {
    // TODO: verify unparse
    std::cout << std::left << std::setw(4) << linenum << ": ";
    genIndent(indent);
    if (this->is_default) {
        std::cout << "case ";
        case_val->unparse();
        std::cout << ": ";
    }
    else {
        std::cout << "default: ";
    }
    stmt->unparse(indent);
} // unparse

bool CaseStmtNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_CaseStmtNode;
} // classof

void CaseStmtNode::setReturnStatus(bool status, Type type) {
    returnStatus = status;
    requiredReturnType = type;
}

void CaseStmtNode::setReturnStatus(bool status) {
    returnStatus = status;
    requiredReturnType = Type(TypeValue::VOID);
}

void CaseStmtNode::checkSemantics() {
    if (!(case_val->isNull())) {
        case_val->checkSemantics();
    }

    // Propagate return type information to case statement
    auto stmtPtr = stmt.get();
    if (llvm::isa<ReturnNode>(stmtPtr)) {
        auto returnPtr = llvm::cast<ReturnNode>(stmtPtr);
        returnPtr->setExpectedType(requiredReturnType);
    }
    else if (llvm::isa<BlockNode>(stmtPtr)) {
        auto caseBlock = llvm::cast<BlockNode>(stmtPtr);
        caseBlock->setReturnStatus(returnStatus, requiredReturnType);
    }
    else if (llvm::isa<SwitchNode>(stmtPtr)) {
        auto switchNode = llvm::cast<SwitchNode>(stmtPtr);
        auto switchStmt = switchNode->body.get();
        if (llvm::isa<ReturnNode>(switchStmt)) {
            ReturnNode* caughtNode = llvm::cast<ReturnNode>(switchStmt);
            caughtNode->setExpectedType(requiredReturnType); // Pass type
        }
        else if (llvm::isa<BlockNode>(switchStmt)) {
            BlockNode* caughtNode = llvm::cast<BlockNode>(switchStmt);
            caughtNode->setReturnStatus(returnStatus, requiredReturnType);
            caughtNode->setGivenScope(true);
        }
    }
    else if (llvm::isa<CaseStmtNode>(stmtPtr)) {
        auto caseNode = llvm::cast<CaseStmtNode>(stmtPtr);
        caseNode->setReturnStatus(returnStatus, requiredReturnType);

        // Also check if this case leads to or is a default case
        leadsToDefault = caseNode->is_default || caseNode->leadsToDefault;
    }

    stmt->checkSemantics();

    // Check if statement returns

    if (llvm::isa<ReturnNode>(stmtPtr)) alwaysReturns = true;
    else if (llvm::isa<BlockNode>(stmtPtr)) {
        auto caseBlock = llvm::cast<BlockNode>(stmtPtr);
        alwaysReturns = caseBlock->alwaysReachesReturn;
    }
    else if (llvm::isa<SwitchNode>(stmtPtr)) {
        auto switchNode = llvm::cast<SwitchNode>(stmtPtr);
        alwaysReturns = switchNode->allBranchesReturn;
    }
    else if (llvm::isa<BreakNode>(stmtPtr)) alwaysReturns = false;
    else alwaysReturns = false;

} // checkSemantics

llvm::Value* CaseStmtNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {

    llvm::ConstantInt* caseVal = nullptr;
    if (!is_default) {
        auto caseExpr = case_val->codegen();
        caseVal = llvm::cast<llvm::ConstantInt>(caseExpr);
    }

    // Create new block for case and save prior insert point
    // (Which should be set to the point after the parent switch statement)
    // for later
    // llvm::BasicBlock* parentBlock = Builder.GetInsertBlock();
    auto parentVal = Builder.GetInsertBlock()->getParent();
    std::string caseBlockName = (is_default) ? "default" : "case";
    llvm::BasicBlock* caseBlock = llvm::BasicBlock::Create(LLVMContext, caseBlockName, parentVal);
    Builder.SetInsertPoint(caseBlock);
    //llvm::Value* caseBodyVal =
    stmt->codegen(parent);

    // Handle fallthrough when statement is just another case
    auto stmtPtr = stmt.get();
    if (llvm::isa<CaseStmtNode>(stmtPtr)) {
        auto stmtBlock = Builder.GetInsertBlock();
        Builder.SetInsertPoint(caseBlock);
        Builder.CreateBr(stmtBlock);
        Builder.SetInsertPoint(stmtBlock);
    }

    // Add case to switch instruction
    auto switchInst = llvm::cast<llvm::SwitchInst>(parent);
    if (is_default) {
        switchInst->setDefaultDest(caseBlock);
    }
    else {
        switchInst->addCase(caseVal, caseBlock);
    }

    // Return builder point after switch instruction
    // Builder.SetInsertPoint(exitBlock);
    return nullptr;

}

// class CaseStmtNode


GotoNode::GotoNode(std::shared_ptr<IdentNode> label, int line, int col) : StmtNode(ASTNodeKind::AST_GotoNode, line, col) {
    this->label = label;
}

void GotoNode::unparse(int indent) {
    // TODO: verify unparse
    std::cout << std::left << std::setw(4) << linenum << ": ";
    genIndent(indent);
    std::cout << "goto ";
    label->unparse();
    std::cout << " ;";
}
bool GotoNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_GotoNode;
} // classof
void GotoNode::checkSemantics() {
    // TODO: implement semantics checks
}
// class GotoNode

// end StmtNode subclasses
