// Controlled gates
// You can add a control qubit to any gate by postfixing with +c(qubit)
%H+C(q2) q1;

// Reverse gate
// Prefixing gates with ~> applies the inverse of what the gate would do
~>%H (%H q2); // same as %I q2;



quantum bool a = 0;
quantum bool b = 1;

// a <> b => a = |1>, b = |0>
// qubits of "a" and "b" are moved to "c" and "d", respectively
// c renames a = |1>, d renames b = |0>
auto [c, d] = a <> b;

quantum bool a[2];
// Allowed; "a" is moved to "b"
quantum bool b[2] = a[1] <> a[2];

// Error; "a" is no longer available
quantum bool c[2] = a[1] <> a[2];




// Rotation gates
// Parameters of gates must be a float
%RX(angle) q1; // Rotate X
%RY(angle) q1; // Rotate Y
%RZ(angle) q1; // Rotate Z


%I q1; // Identity (no-op)
%X q1; // Pauli X gate
%Y q1; // Pauli Y gate
%Z q1; // Pauli Z gate
%H q1; // Hadamard



quantum bool a = 0;
quantum bool b = 0;
quantum bool c = %H a; // Allowed: "a" is moved to "c"
quantum bool d = %H a; // Error

quantum int q1;
%X (%H q1);
int a = %M q1; // Measurement
int b = %H q1; // Error

quantum int a = 2;   // a = |00...0010>
quantum int b;       // b = |00...0000>
quantum bool c;      // c = |0>
b = 5;               // Error
int n = 3;
quantum bool d[3] = {0, 1, 0};    // Okay; d = {|0>, |1>, |0>}
quantum bool e[n];    // Error


%I q1; // Identity (no-op)
%X q1; // Pauli X gate
%Y q1; // Pauli Y gate
%Z q1; // Pauli Z gate
%H q1; // Hadamard

q1 ^ q2; // Controlled NOT (CNOT)
// Swap gates
// Parameters of gates must be float
q1 <P>(angle) q2; // General swap; parameterized by an angle
q1 <> q2; // Classical swap; angle = 0
q1 <I> q2; // Imaginary swap; angle = pi/2



/*
 Note: this program isn't neccesarily semantically correct, just grammatically
 Quantum computer is assumed to have both classical and quantum
 memory
*/


/*
 quantum keyword used as a storage class identifier to indicate sections of
 code for the quantum computer.
 The quantum keyword (for now) is limited to functions and variables of builtin
 types.

 For quantum functions,
 1) Classical value arguments can be passed by value or reference.
 2) Quantum value arguments MUST be passed by reference.
 3) Quantum functions which contain quantum value arguments MAY NOT be
    called by classical code.
 4) A quantum function callable from classical code MUST contain only
    classical arguments passed by value.
 5) A quantum function callable from classical code MAY return a value.
 6) Quantum functions MUST NOT be recursive.
*/









quantum void basic_function(int value1, quantum int &value2) {
    return;
}


quantum int hello_quantum() {

    /*
      Standard types supported in quantum.
      Standard operations are supported for classical types.
    */
    int a, b, c;

    // Pointer types are NOT allowed inside quantum functions.
    // likewise, getting references to other variables is disallowed except when
    // used in a quantum function call.
    int * pointer;
    pointer = &b;

    /*
     PI will be reserved as a constant, since quantum gates
     often need to define angles as parameters of gates
    */
    float angle = PI/2;

    /*
     Quantum storage classifier added to specify that quantum memory (qubits)
     is to be used; otherwise identical to classic types.
     Floating point values are disallowed.
     Quantum variables can only appear inside quantum functions.
     Standard operators not allowed on quantum variables; only quantum operators.
    */
    quantum int q1, q2, result = 0;

    // Assignment
    a = b; // Allowed; assignment between classical memory acts like normal
    q1 = a; // Allowed; assignment from classical to quantum memory simply sets qubits to corresponding base states.
    a = q1; // Denied; assignment from quantum to classical memory needs to have the qubit measured first.
    a = %M q1; // Allowed; use of measurement operator allows for
    q2 = q1; // Denied; cannot copy arbitray quantum states

    // if statements can be controlled with both classical bits and qubits,
    // but qubits must be measured before
    if (%M (q1)) { // Need to measure qubits if using for conditonal statements
        q2 = 0;
    }
    else {
        q2 = 1;
    }

    // loops are allowed, but must use classical types to control them.
    //for (int i = 0; i < a; ++i) {
    //}
    //int i = 0;
    while (i < a) {
        ++i;
    }

    // Other gates (and controlled versions)

    // Classical gates
    ~q1;     // NOT (also called X gate);
    // Implictly CNOT; q2 becomes q1 XOR q2; expression returns q2
    q1 ^ q2;
    // Implictly Toffoli (CCNOT); result becomes result XOR q1 AND q2;
    // expression returns result
    // the third argument is required when using quantum variables
    q1 & q2 : result;

    // Implicitly Toffoli + NOT gates; result becomes result XOR q1 OR q2;
    // expression returns result
    // the third argument is required when using quantum variables
    q1 | q2 : result;

    %I q1; // Identity gates (does not change qubit)

    %Y q1; // Pauli Y gate
    %Z q1; // Pauli Z gate
    %H q1; // Hadamard

    // Phase gates
    // Parameters of gates must be a float
    %PHASE(angle) q1; // General phase gates, parametrized by an angle
    %S q1; // phase gate with angle = pi/2
    %T q1; // phase gate with angle = pi/4

    // Rotation gates
    // Parameters of gates must be a float
    %RX(angle) q1;
    %RY(angle) q1;
    %RZ(angle) q1;

    // Swap gates
    // Parameters of gates must be a float
    q1 <P>(angle) q2; // General swap; parametrized by an angle
    q1 <> q2; // Classical swap; angle = 0
    q1 <I> q2; // Imaginary swap; angle = pi/2

    // Controlled gates
    // You can add a control qubit to any gate by postfixing with +c(qubit)
    %H+C(q2) q1;

    // Reverse gate
    // Prefixing gate with ~> does the inverse of what the gate would do
    ~>%H (%H q2); // same as %I q2;

    // Calling quantum functions
    // Note that quantum types must be passed by reference.
    basic_function(a, &q1);

    /* Return value
    Only quantum functions that are called by classical code can return a value.
    The result will be whatever the value is when measured.
    */
    return result;
}

void main() {

    // Callable quantum functions can return a type
    int test_int = hello_quantum();
}
